# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from . import bwepdio as bwm
from . import cpmemio as cpm
from . import dwepdio as dwm
from . import modcurpl as cp


def _ptr2epd(ptr):
    if c.IsEUDVariable(ptr):
        epd, subp = c.f_div(ptr + (-0x58A364), 4)
    else:
        epd, subp = divmod(ptr + (-0x58A364), 4)
    return epd, subp


def f_dwwrite(ptr, dw):
    if isinstance(ptr, int) and ptr % 4 == 0:
        dwm.f_dwwrite_epd(ut.EPD(ptr), dw)
    else:
        chars = dwm.f_dwbreak(dw)[2:]
        from .rwcommon import bw1

        _bw = bw1
        _bw.seekoffset(ptr)
        _bw.writebyte(chars[0])
        _bw.writebyte(chars[1])
        _bw.writebyte(chars[2])
        _bw.writebyte(chars[3])


def f_wwrite(ptr, w):
    epd, subp = _ptr2epd(ptr)
    bwm.f_wwrite_epd(epd, subp, w)


def f_bwrite(ptr, b):
    epd, subp = _ptr2epd(ptr)
    bwm.f_bwrite_epd(epd, subp, b)


# -----------------------------


@c.EUDFunc
def _dwread(ptr):
    ptr -= 0x58A364
    dw, subp = c.EUDCreateVariables(2)
    c.f_div(ptr, 4, ret=[ut.EPD(0x6509B0), subp])
    dw << 0  # TODO: merge this action to f_div call trigger
    cs.EUDSwitch(subp)

    # Case 0
    if cs.EUDSwitchCase()(0):
        cpm.f_dwread_cp(0, ret=[dw])
        cs.EUDBreak()

    # Else → Complex
    for i in ut._rand_lst(range(1, 4)):
        cs.EUDSwitchCase()(i)

        for j in ut._rand_lst(range(8 * i, 32)):
            c.RawTrigger(
                conditions=c.DeathsX(cp.CP, c.AtLeast, 1, 0, 2**j),
                actions=dw.AddNumber(2 ** (j - 8 * i)),
            )

        c.SeqCompute([(ut.EPD(0x6509B0), c.Add, 1)])

        for j in ut._rand_lst(range(8 * i)):
            c.RawTrigger(
                conditions=c.DeathsX(cp.CP, c.AtLeast, 1, 0, 2**j),
                actions=dw.AddNumber(2 ** (j + 32 - 8 * i)),
            )

        cs.EUDBreak()

    cs.EUDEndSwitch()
    cp.f_setcurpl2cpcache()
    return dw


def f_dwread(ptr, **kwargs):
    if (isinstance(ptr, int) and ptr % 4 == 0) or (
        not c.IsEUDVariable(ptr)
        and isinstance(ptr, (c.RlocInt_C, c.ConstExpr))  # noqa: UP038
        and ptr._is_aligned_ptr()
    ):
        return dwm.f_dwread_epd(ut.EPD(ptr), **kwargs)
    else:
        return _dwread(ptr, **kwargs)


def f_wread(ptr, **kwargs):
    return bwm.f_wread_epd(*_ptr2epd(ptr), **kwargs)


def f_bread(ptr, **kwargs):
    return bwm.f_bread_epd(*_ptr2epd(ptr), **kwargs)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...localize import _


def EUDBinaryMax(cond, minv=0, maxv=0xFFFFFFFF) -> c.EUDVariable:  # noqa: N802
    """Find maximum x satisfying cond(x) using binary search

    :param cond: Test condition
    :param minv: Minimum value in domain
    :param maxv: Maximum value in domain

    Cond should be binary classifier, meaning that for some N
        for all x > N, cond(x) is false.
        for all x <= N, cond(x) is true
    Then EUDBinaryMin will find such N.

    .. note:: If none of the value satisfies condition, then this
        function will return maxv.
    """
    if isinstance(minv, int) and isinstance(maxv, int):
        ut.ep_assert(0 <= minv <= maxv, _("maxv must be greater than minv"))
        r = maxv - minv
    else:
        r = None

    x = c.EUDVariable()
    nth = 0
    for i in range(31, -1, -1):
        if r and 2**i > r:
            continue

        if nth == 0:
            x << minv + 2**i
            if cs.EUDIfNot()(cond(x)):
                cs.DoActions(x.SubtractNumber(2**i))
            cs.EUDEndIf()
        else:
            x += 2**i
            if cs.EUDIfNot()([x <= maxv, cond(x)]):
                cs.DoActions(x.SubtractNumber(2**i))
            cs.EUDEndIf()
        nth += 1

    return x


def EUDBinaryMin(cond, minv=0, maxv=0xFFFFFFFF) -> c.EUDVariable:  # noqa: N802
    """Find minimum x satisfying cond(x) using binary search

    :param cond: Test condition
    :param minv: Minimum value in domain
    :param maxv: Maximum value in domain

    Cond should be binary classifier, meaning that for some N
        for all x < N, cond(x) is false.
        for all x >= N, cond(x) is true
    Then EUDBinaryMin will find such N

    .. note:: If none of the value satisfies condition, then this
        function will return maxv.
    """
    if isinstance(minv, int) and isinstance(maxv, int):
        ut.ep_assert(0 <= minv <= maxv, _("maxv must be greater than minv"))
        r = maxv - minv
    else:
        r = None

    x = c.EUDVariable()
    nth = 0
    for i in range(31, -1, -1):
        if r and 2**i > r:
            continue

        if nth == 0:
            x << maxv - 2**i
            if cs.EUDIfNot()(cond(x)):
                x += 2**i
            cs.EUDEndIf()
        else:
            if cs.EUDIf()(x >= 2**i):
                cs.DoActions(x.SubtractNumber(2**i))
                if cs.EUDIfNot()([x >= minv, cond(x)]):
                    x += 2**i
                cs.EUDEndIf()
            cs.EUDEndIf()
        nth += 1

    return x



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .runtrigtrg import EUDLoopTrigger
from .trigtrg import (
    Accumulate,
    Action,
    Always,
    Briefing,
    Bring,
    CenterView,
    Command,
    CommandLeast,
    CommandLeastAt,
    CommandMost,
    CommandMostAt,
    Comment,
    Condition,
    CountdownTimer,
    CreateUnit,
    CreateUnitWithProperties,
    Deaths,
    DeathsX,
    Defeat,
    DisplayText,
    Draw,
    ElapsedTime,
    GiveUnits,
    HighestScore,
    Kills,
    KillUnit,
    KillUnitAt,
    LeaderBoardComputerPlayers,
    LeaderBoardControl,
    LeaderBoardControlAt,
    LeaderBoardGoalControl,
    LeaderBoardGoalControlAt,
    LeaderBoardGoalKills,
    LeaderBoardGoalResources,
    LeaderBoardGoalScore,
    LeaderBoardGreed,
    LeaderBoardKills,
    LeaderBoardResources,
    LeaderBoardScore,
    LeastKills,
    LeastResources,
    LowestScore,
    Memory,
    MemoryX,
    MinimapPing,
    ModifyUnitEnergy,
    ModifyUnitHangarCount,
    ModifyUnitHitPoints,
    ModifyUnitResourceAmount,
    ModifyUnitShields,
    MostKills,
    MostResources,
    MoveLocation,
    MoveUnit,
    MuteUnitSpeech,
    Never,
    NoAction,
    NoCondition,
    Opponents,
    Order,
    PauseGame,
    PauseTimer,
    PlayWAV,
    PreserveTrigger,
    RemoveUnit,
    RemoveUnitAt,
    RunAIScript,
    RunAIScriptAt,
    Score,
    SetAllianceStatus,
    SetCountdownTimer,
    SetDeaths,
    SetDeathsX,
    SetDoodadState,
    SetInvincibility,
    SetMemory,
    SetMemoryX,
    SetMissionObjectives,
    SetNextScenario,
    SetResources,
    SetScore,
    SetSwitch,
    Switch,
    TalkingPortrait,
    Transmission,
    Trigger,
    UnMuteUnitSpeech,
    UnpauseGame,
    UnpauseTimer,
    Victory,
    Wait,
)

__all__ = [
    "EUDLoopTrigger",
    "Accumulate",
    "Action",
    "Always",
    "Briefing",
    "Bring",
    "CenterView",
    "Command",
    "CommandLeast",
    "CommandLeastAt",
    "CommandMost",
    "CommandMostAt",
    "Comment",
    "Condition",
    "CountdownTimer",
    "CreateUnit",
    "CreateUnitWithProperties",
    "Deaths",
    "DeathsX",
    "Defeat",
    "DisplayText",
    "Draw",
    "ElapsedTime",
    "GiveUnits",
    "HighestScore",
    "Kills",
    "KillUnit",
    "KillUnitAt",
    "LeaderBoardComputerPlayers",
    "LeaderBoardControl",
    "LeaderBoardControlAt",
    "LeaderBoardGoalControl",
    "LeaderBoardGoalControlAt",
    "LeaderBoardGoalKills",
    "LeaderBoardGoalResources",
    "LeaderBoardGoalScore",
    "LeaderBoardGreed",
    "LeaderBoardKills",
    "LeaderBoardResources",
    "LeaderBoardScore",
    "LeastKills",
    "LeastResources",
    "LowestScore",
    "Memory",
    "MemoryX",
    "MinimapPing",
    "ModifyUnitEnergy",
    "ModifyUnitHangarCount",
    "ModifyUnitHitPoints",
    "ModifyUnitResourceAmount",
    "ModifyUnitShields",
    "MostKills",
    "MostResources",
    "MoveLocation",
    "MoveUnit",
    "MuteUnitSpeech",
    "Never",
    "NoAction",
    "NoCondition",
    "Opponents",
    "Order",
    "PauseGame",
    "PauseTimer",
    "PlayWAV",
    "PreserveTrigger",
    "RemoveUnit",
    "RemoveUnitAt",
    "RunAIScript",
    "RunAIScriptAt",
    "Score",
    "SetAllianceStatus",
    "SetCountdownTimer",
    "SetDeaths",
    "SetDeathsX",
    "SetDoodadState",
    "SetInvincibility",
    "SetMemory",
    "SetMemoryX",
    "SetMissionObjectives",
    "SetNextScenario",
    "SetResources",
    "SetScore",
    "SetSwitch",
    "Switch",
    "TalkingPortrait",
    "Transmission",
    "Trigger",
    "UnMuteUnitSpeech",
    "UnpauseGame",
    "UnpauseTimer",
    "Victory",
    "Wait",
]



# Copyright 2022 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from types import ModuleType

from ..collections import EUDArray
from ..core import (
    ConstExpr,
    EUDVariable,
    EUDVArray,
    Forward,
    IsEUDVariable,
    NextTrigger,
    PopTriggerScope,
    PushTriggerScope,
    SeqCompute,
    SetNextTrigger,
    SetTo,
    f_bitlshift,
)
from ..ctrlstru import EUDElse, EUDEndIf, EUDIf
from ..maprw import EUDOnStart
from ..memio import f_setcurpl2cpcache
from ..utils import (
    EPD,
    ExprProxy,
    FlattenList,
    List2Assignable,
    ep_assert,
    ep_warn,
    isUnproxyInstance,
)


def _RELIMP(path, mod_name):  # noqa: N802
    """relative path import"""
    import importlib
    import inspect
    import pathlib
    import sys

    abs_path = pathlib.Path(inspect.getabsfile(inspect.currentframe().f_back))
    pathsplit = path.split(".")
    if len(path) + 1 == len(pathsplit):  # path == '.' * len(path)
        pathsplit = pathsplit[1:]
    for s in pathsplit:
        if s == "":
            abs_path = abs_path.parent
        else:
            abs_path = abs_path / s

    abs_path = abs_path / mod_name
    abs_import = None
    # test if relative import can be converted to absolute import
    for parent in sys.path:
        try:
            abs_import = abs_path.relative_to(parent)
        except ValueError:
            continue
        else:
            break
    if abs_import is None:
        raise ImportError("attempted relative import beyond top-level package")
    try:
        return importlib.import_module(str(abs_import).replace("\\", "."))
    except ImportError:
        module = importlib.import_module(str(abs_import.parent).replace("\\", "."))
        try:
            return getattr(module, mod_name)
        except AttributeError as err:
            raise ImportError from err


def _TYGV(types, expr_list_gen):  # noqa: N802
    PushTriggerScope()
    init_start = NextTrigger()
    values = expr_list_gen()
    var_list = []
    for ty, value in zip(types, values):
        is_eudvar = IsEUDVariable(value)
        is_untyped = ty is None or issubclass(ty, EUDVariable)
        if not is_untyped and not (type(value) is int and value == 0):  # noqa: E721
            value = ty.cast(value)
        if not is_eudvar:
            value = EUDVariable(value)
            if not is_untyped:
                value = ty.cast(value)
        var_list.append(value)

    if init_start.IsSet():
        nptr = Forward()
        SetNextTrigger(nptr)

        def _init_global_var() -> None:
            nonlocal nptr
            SetNextTrigger(init_start)
            nptr << NextTrigger()

        if init_start.IsSet():
            EUDOnStart(_init_global_var)

    PopTriggerScope()

    return List2Assignable(var_list)


def _TYSV(types, vs):  # noqa: N802
    var_list = []
    for ty, v in zip(types, vs):
        is_untyped = ty is None or issubclass(ty, EUDVariable)
        if not is_untyped and not (type(v) is int and v == 0):  # noqa: E721
            v = ty.cast(v)
        value = EUDVariable(v)
        if not is_untyped:
            value = ty.cast(value)
        var_list.append(value)
    return List2Assignable(var_list)


def _TYLV(types, vs):  # noqa: N802
    from ..core.variable.eudv import _yield_and_check_rvalue

    ret, ops = [], []
    for ty, v_and_is_rvalue in zip(types, _yield_and_check_rvalue(vs, 1)):
        is_untyped = ty is None or issubclass(ty, EUDVariable)
        v, is_rvalue = v_and_is_rvalue
        is_eudvar = IsEUDVariable(v)
        if is_eudvar and is_rvalue:
            nv = v.makeL()
        else:
            if not is_untyped and not (type(v) is int and v == 0):  # noqa: E721
                v = ty.cast(v)
            nv = EUDVariable()
            ops.append((nv, SetTo, v))
        if not is_untyped:
            nv = ty.cast(nv)
        ret.append(nv)
    if ops:
        SeqCompute(ops)
    return List2Assignable(ret)


def _CGFW(exprf, retn):  # noqa: N802
    PushTriggerScope()
    start = NextTrigger()
    try:
        rets = exprf()
    except NameError:
        rets = [ExprProxy(None) for _ in range(retn)]

        def _init_global_const():
            vals = exprf()
            for ret, val in zip(rets, vals):
                ret._value = val

        EUDOnStart(_init_global_const)
    if start.IsSet():
        end = Forward()
        SetNextTrigger(end)

        def _init():
            nonlocal end
            SetNextTrigger(start)
            end << NextTrigger()

        EUDOnStart(_init)

    PopTriggerScope()

    return rets


def _ARR(items):  # EUDArray initialization  # noqa: N802
    k = EUDArray(len(items))
    for i, item in enumerate(items):
        k[i] = item
    return k


def _VARR(items):  # EUDVArray initialization  # noqa: N802
    k = EUDVArray(len(items))()
    for i, item in enumerate(items):
        k[i] = item
    return k


def _SRET(v, klist):  # noqa: N802
    return List2Assignable([v[k] for k in klist])


def _SV(d_list, s_list):  # noqa: N802
    [d << s for d, s in zip(FlattenList(d_list), FlattenList(s_list))]


class _ATTW:  # attribute write
    def __init__(self, obj, attr_name):
        self.obj = obj
        self.attr_name = attr_name

    def __lshift__(self, r):
        if isinstance(self.obj, ModuleType):
            ov = getattr(self.obj, self.attr_name)
            if IsEUDVariable(ov):
                ov << r
                return
            ep_warn("Try to shadow module variable")
        setattr(self.obj, self.attr_name, r)

    def __iadd__(self, v):
        try:
            self.obj.iaddattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov += v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __isub__(self, v):
        try:
            self.obj.isubattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov -= v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __imul__(self, v):
        try:
            self.obj.imulattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov *= v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __ifloordiv__(self, v):
        try:
            self.obj.ifloordivattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov //= v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __imod__(self, v):
        try:
            self.obj.imodattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov %= v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __ilshift__(self, v):
        try:
            self.obj.ilshiftattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov <<= v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __irshift__(self, v):
        try:
            self.obj.irshiftattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov >>= v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __ipow__(self, v):
        try:
            self.obj.ipowattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov **= v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __iand__(self, v):
        try:
            self.obj.iandattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov &= v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __ior__(self, v):
        try:
            self.obj.iorattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov |= v
            setattr(self.obj, self.attr_name, ov)
        return self

    def __ixor__(self, v):
        try:
            self.obj.ixorattr(self.attr_name, v)
        except AttributeError:
            ov = getattr(self.obj, self.attr_name)
            ov ^= v
            setattr(self.obj, self.attr_name, ov)
        return self


class _ARRW:  # array write
    def __init__(self, obj, index):
        self.obj = obj
        self.index = index

    def __lshift__(self, r):
        isUnproxyInstance
        if not IsEUDVariable(self.obj) and not isUnproxyInstance(
            self.obj, ConstExpr
        ):
            # maybe Python collections
            ov = self.obj[self.index]
            if IsEUDVariable(ov):
                ov << r
                return
        self.obj[self.index] = r

    def __iadd__(self, v):
        try:
            self.obj.iadditem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov += v
            self.obj[self.index] = ov
        return self

    def __isub__(self, v):
        try:
            self.obj.isubitem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov -= v
            self.obj[self.index] = ov
        return self

    def __imul__(self, v):
        try:
            self.obj.imulitem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov *= v
            self.obj[self.index] = ov
        return self

    def __ifloordiv__(self, v):
        try:
            self.obj.ifloordivitem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov //= v
            self.obj[self.index] = ov
        return self

    def __imod__(self, v):
        try:
            self.obj.imoditem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov %= v
            self.obj[self.index] = ov
        return self

    def __ilshift__(self, v):
        try:
            self.obj.ilshiftitem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov <<= v
            self.obj[self.index] = ov
        return self

    def __irshift__(self, v):
        try:
            self.obj.irshiftitem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov >>= v
            self.obj[self.index] = ov
        return self

    def __ipow__(self, v):
        try:
            self.obj.ipowitem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov **= v
            self.obj[self.index] = ov
        return self

    def __iand__(self, v):
        try:
            self.obj.ianditem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov &= v
            self.obj[self.index] = ov
        return self

    def __ior__(self, v):
        try:
            self.obj.ioritem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov |= v
            self.obj[self.index] = ov
        return self

    def __ixor__(self, v):
        try:
            self.obj.ixoritem(self.index, v)
        except AttributeError:
            ov = self.obj[self.index]
            ov ^= v
            self.obj[self.index] = ov
        return self


class _ATTC:  # attribute comparison
    def __init__(self, obj, attr_name):
        self.obj = obj
        self.attr_name = attr_name

    def __eq__(self, k):
        try:
            return self.obj.eqattr(self.attr_name, k)
        except AttributeError:
            return getattr(self.obj, self.attr_name) == k

    def __ne__(self, k):
        try:
            return self.obj.neattr(self.attr_name, k)
        except AttributeError:
            return getattr(self.obj, self.attr_name) != k

    def __le__(self, k):
        try:
            return self.obj.leattr(self.attr_name, k)
        except AttributeError:
            return getattr(self.obj, self.attr_name) <= k

    def __lt__(self, k):
        try:
            return self.obj.ltattr(self.attr_name, k)
        except AttributeError:
            return getattr(self.obj, self.attr_name) < k

    def __ge__(self, k):
        try:
            return self.obj.geattr(self.attr_name, k)
        except AttributeError:
            return getattr(self.obj, self.attr_name) >= k

    def __gt__(self, k):
        try:
            return self.obj.gtattr(self.attr_name, k)
        except AttributeError:
            return getattr(self.obj, self.attr_name) > k


class _ARRC:  # array comparison
    def __init__(self, obj, index):
        self.obj = obj
        self.index = index

    def __eq__(self, k):
        try:
            return self.obj.eqitem(self.index, k)
        except AttributeError:
            return self.obj[self.index] == k

    def __ne__(self, k):
        try:
            return self.obj.neitem(self.index, k)
        except AttributeError:
            return self.obj[self.index] != k

    def __le__(self, k):
        try:
            return self.obj.leitem(self.index, k)
        except AttributeError:
            return self.obj[self.index] <= k

    def __lt__(self, k):
        try:
            return self.obj.ltitem(self.index, k)
        except AttributeError:
            return self.obj[self.index] < k

    def __ge__(self, k):
        try:
            return self.obj.geitem(self.index, k)
        except AttributeError:
            return self.obj[self.index] >= k

    def __gt__(self, k):
        try:
            return self.obj.gtitem(self.index, k)
        except AttributeError:
            return self.obj[self.index] > k


def _L2V(logic):  # noqa: N802, logic to value
    ret = EUDVariable()
    if EUDIf()(logic):
        ret << 1
    if EUDElse()():
        ret << 0
    EUDEndIf()
    return ret


def _LSH(lhs, r):  # noqa: N802
    if IsEUDVariable(lhs):
        return f_bitlshift(lhs, r)
    else:
        return lhs << r


def _ALL(actions):  # noqa: N802
    from ..trigger.tpatcher import actpt, apply_patch_table

    ep_assert(len(actions) == 2, f"Invalid AllPlayers action: {actions}")
    apply_patch_table(EPD(actions[1]), actions[1], actpt)
    f_setcurpl2cpcache([], actions)



from helper import *


@EUDFunc
def callonly():
    pass


def empty():
    pass


@TestInstance
def test_perfbasic():
    test_perf("Basic looping", lambda: None, 100 * perf_basecount)
    test_perf("Basic function call", EUDFunc(lambda: None), 100 * perf_basecount)

    a, b = EUDCreateVariables(2)
    SetVariables([a, b], [f_dwrand(), f_rand()])

    # Comparison of various functions
    test_perf("Addition", lambda: a + b, perf_basecount)
    test_perf("Subtraction", lambda: a - b, perf_basecount)
    test_perf("Multiplication", lambda: a * b, perf_basecount)
    test_perf("Division", lambda: a // b, perf_basecount)



from typing import Literal, TypeAlias

DefWeaponDict: dict[str, int] = {
    "Gauss Rifle": 0,
    "Gauss Rifle (Raynor)": 1,
    "C-10 Concussion Rifle": 2,
    "C-10 Concussion Rifle (Kerrigan)": 3,
    "Fragmentation Grenade": 4,
    "Fragmentation Grenade (Raynor)": 5,
    "Spider Mines": 6,
    "Twin Autocannons": 7,
    "Hellfire Missile Pack": 8,
    "Twin Autocannons (Schezar)": 9,
    "Hellfire Missile Pack (Schezar)": 10,
    "Arclite Cannon": 11,
    "Arclite Cannon (Duke)": 12,
    "Fusion Cutter": 13,
    "Fusion Cutter (Harvest)": 14,
    "Gemini Missiles": 15,
    "Burst Lasers": 16,
    "Gemini Missiles (Kazansky)": 17,
    "Burst Lasers (Kazansky)": 18,
    "ATS Laser Battery": 19,
    "ATA Laser Battery": 20,
    "ATS Laser Battery (Norad II+Mengsk+DuGalle)": 21,
    "ATA Laser Battery (Norad II+Mengsk+DuGalle)": 22,
    "ATS Laser Battery (Hyperion)": 23,
    "ATA Laser Battery (Hyperion)": 24,
    "Flame Thrower": 25,
    "Flame Thrower (Montag)": 26,
    "Arclite Shock Cannon": 27,
    "Arclite Shock Cannon (Duke)": 28,
    "Longbolt Missiles": 29,
    "Yamato Gun": 30,
    "Nuclear Missile": 31,
    "Lockdown": 32,
    "EMP Shockwave": 33,
    "Irradiate": 34,
    "Claws": 35,
    "Claws (Devouring One)": 36,
    "Claws (Infested Kerrigan)": 37,
    "Needle Spines": 38,
    "Needle Spines (Hunter Killer)": 39,
    "Kaiser Blades": 40,
    "Kaiser Blades (Torrasque)": 41,
    "Toxic Spores": 42,
    "Spines": 43,
    "Spines (Harvest)": 44,
    "Acid Spray": 45,
    "Acid Spore": 46,
    "Acid Spore (Kukulza)": 47,
    "Glave Wurm": 48,
    "Glave Wurm (Kukulza)": 49,
    "Venom": 50,
    "Venom (Unclean One)": 51,
    "Seeker Spores": 52,
    "Subterranean Tentacle": 53,
    "Suicide (Infested Terran)": 54,
    "Suicide (Scourge)": 55,
    "Parasite": 56,
    "Spawn Broodlings": 57,
    "Ensnare": 58,
    "Dark Swarm": 59,
    "Plague": 60,
    "Consume": 61,
    "Particle Beam": 62,
    "Particle Beam (Harvest)": 63,
    "Psi Blades": 64,
    "Psi Blades (Fenix)": 65,
    "Phase Disruptor": 66,
    "Phase Disruptor (Fenix)": 67,
    "Psi Assault": 68,
    "Psi Assault (Tassadar)": 69,
    "Psionic Shockwave": 70,
    "Psionic Shockwave (Tassadar/Zeratul)": 71,
    "Unused72": 72,
    "Dual Photon Blasters": 73,
    "Anti-matter Missiles": 74,
    "Dual Photon Blasters (Mojo)": 75,
    "Anti-matter Missiles (Mojo)": 76,
    "Phase Disruptor Cannon": 77,
    "Phase Disruptor Cannon (Danimoth)": 78,
    "Pulse Cannon": 79,
    "STS Photon Cannon": 80,
    "STA Photon Cannon": 81,
    "Scarab": 82,
    "Stasis Field": 83,
    "Psi Storm": 84,
    "Warp Blades (Zeratul)": 85,
    "Warp Blades (Hero)": 86,
    "Missiles": 87,
    "Laser Battery1": 88,
    "Tormentor Missiles": 89,
    "Bombs": 90,
    "Raider Gun": 91,
    "Laser Battery2": 92,
    "Laser Battery3": 93,
    "Dual Photon Blasters (Unused)": 94,
    "Flechette Grenade": 95,
    "Twin Autocannons (Floor Trap)": 96,
    "Hellfire Missile Pack (Wall Trap)": 97,
    "Flame Thrower (Wall Trap)": 98,
    "Hellfire Missile Pack (Floor Trap)": 99,
    "Neutron Flare": 100,
    "Disruption Web": 101,
    "Restoration": 102,
    "Halo Rockets": 103,
    "Corrosive Acid": 104,
    "Mind Control": 105,
    "Feedback": 106,
    "Optical Flare": 107,
    "Maelstrom": 108,
    "Subterranean Spines": 109,
    "Gauss Rifle0": 110,
    "Unused110": 110,
    "Warp Blades": 111,
    "C-10 Concussion Rifle (Samir Duran)": 112,
    "C-10 Concussion Rifle (Infested Duran)": 113,
    "Dual Photon Blasters (Artanis)": 114,
    "Anti-matter Missiles (Artanis)": 115,
    "C-10 Concussion Rifle (Stukov)": 116,
    "Gauss Rifle1": 117,
    "Gauss Rifle2": 118,
    "Gauss Rifle3": 119,
    "Gauss Rifle4": 120,
    "Gauss Rifle5": 121,
    "Gauss Rifle6": 122,
    "Gauss Rifle7": 123,
    "Gauss Rifle8": 124,
    "Gauss Rifle9": 125,
    "Gauss Rifle10": 126,
    "Gauss Rifle11": 127,
    "Gauss Rifle12": 128,
    "Gauss Rifle13": 129,
    "Unused117": 117,
    "Unused118": 118,
    "Unused119": 119,
    "Unused120": 120,
    "Unused121": 121,
    "Unused122": 122,
    "Unused123": 123,
    "Unused124": 124,
    "Unused125": 125,
    "Unused126": 126,
    "Unused127": 127,
    "Unused128": 128,
    "Unused129": 129,
    "None": 130,
}

DefaultWeapon: TypeAlias = Literal[
    "Gauss Rifle",
    "Gauss Rifle (Raynor)",
    "C-10 Concussion Rifle",
    "C-10 Concussion Rifle (Kerrigan)",
    "Fragmentation Grenade",
    "Fragmentation Grenade (Raynor)",
    "Spider Mines",
    "Twin Autocannons",
    "Hellfire Missile Pack",
    "Twin Autocannons (Schezar)",
    "Hellfire Missile Pack (Schezar)",
    "Arclite Cannon",
    "Arclite Cannon (Duke)",
    "Fusion Cutter",
    "Fusion Cutter (Harvest)",
    "Gemini Missiles",
    "Burst Lasers",
    "Gemini Missiles (Kazansky)",
    "Burst Lasers (Kazansky)",
    "ATS Laser Battery",
    "ATA Laser Battery",
    "ATS Laser Battery (Norad II+Mengsk+DuGalle)",
    "ATA Laser Battery (Norad II+Mengsk+DuGalle)",
    "ATS Laser Battery (Hyperion)",
    "ATA Laser Battery (Hyperion)",
    "Flame Thrower",
    "Flame Thrower (Montag)",
    "Arclite Shock Cannon",
    "Arclite Shock Cannon (Duke)",
    "Longbolt Missiles",
    "Yamato Gun",
    "Nuclear Missile",
    "Lockdown",
    "EMP Shockwave",
    "Irradiate",
    "Claws",
    "Claws (Devouring One)",
    "Claws (Infested Kerrigan)",
    "Needle Spines",
    "Needle Spines (Hunter Killer)",
    "Kaiser Blades",
    "Kaiser Blades (Torrasque)",
    "Toxic Spores",
    "Spines",
    "Spines (Harvest)",
    "Acid Spray",
    "Acid Spore",
    "Acid Spore (Kukulza)",
    "Glave Wurm",
    "Glave Wurm (Kukulza)",
    "Venom",
    "Venom (Unclean One)",
    "Seeker Spores",
    "Subterranean Tentacle",
    "Suicide (Infested Terran)",
    "Suicide (Scourge)",
    "Parasite",
    "Spawn Broodlings",
    "Ensnare",
    "Dark Swarm",
    "Plague",
    "Consume",
    "Particle Beam",
    "Particle Beam (Harvest)",
    "Psi Blades",
    "Psi Blades (Fenix)",
    "Phase Disruptor",
    "Phase Disruptor (Fenix)",
    "Psi Assault",
    "Psi Assault (Tassadar)",
    "Psionic Shockwave",
    "Psionic Shockwave (Tassadar/Zeratul)",
    "Unused72",
    "Dual Photon Blasters",
    "Anti-matter Missiles",
    "Dual Photon Blasters (Mojo)",
    "Anti-matter Missiles (Mojo)",
    "Phase Disruptor Cannon",
    "Phase Disruptor Cannon (Danimoth)",
    "Pulse Cannon",
    "STS Photon Cannon",
    "STA Photon Cannon",
    "Scarab",
    "Stasis Field",
    "Psi Storm",
    "Warp Blades (Zeratul)",
    "Warp Blades (Hero)",
    "Missiles",
    "Laser Battery1",
    "Tormentor Missiles",
    "Bombs",
    "Raider Gun",
    "Laser Battery2",
    "Laser Battery3",
    "Dual Photon Blasters (Unused)",
    "Flechette Grenade",
    "Twin Autocannons (Floor Trap)",
    "Hellfire Missile Pack (Wall Trap)",
    "Flame Thrower (Wall Trap)",
    "Hellfire Missile Pack (Floor Trap)",
    "Neutron Flare",
    "Disruption Web",
    "Restoration",
    "Halo Rockets",
    "Corrosive Acid",
    "Mind Control",
    "Feedback",
    "Optical Flare",
    "Maelstrom",
    "Subterranean Spines",
    "Unused110",
    "Warp Blades",
    "C-10 Concussion Rifle (Samir Duran)",
    "C-10 Concussion Rifle (Infested Duran)",
    "Dual Photon Blasters (Artanis)",
    "Anti-matter Missiles (Artanis)",
    "C-10 Concussion Rifle (Stukov)",
    "Unused117",
    "Unused118",
    "Unused119",
    "Unused120",
    "Unused121",
    "Unused122",
    "Unused123",
    "Unused124",
    "Unused125",
    "Unused126",
    "Unused127",
    "Unused128",
    "Unused129",
    "None",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Iterable

from .. import core as c
from .. import utils as ut
from ..localize import _
from .tpatcher import patch_action, patch_condition

Conditions = (
    c.Condition | bool | Iterable[c.Condition | bool | Iterable | None] | None
)
Actions = c.Action | Iterable[c.Action | Iterable | None] | None


def Trigger(  # noqa: N802
    conditions: Conditions = None,
    actions: Actions = None,
    preserved: bool = True,
) -> tuple[c.Forward, c.RawTrigger]:
    """General easy-to-use trigger

    :param conditions: List of conditions. If there are none, trigger will
        always execute.
    :param actions: List of actions. If there are none, trigger will have no
        actions.
    :param preserved: Is trigger preserved? True by default.

    .. note::
        This is 'extended' trigger. All conditions and variables can contain
        `EUDVariable` object, and there may be more than 16 conditions and 64
        actions. Trigger internally uses `RawTrigger`.
    """

    ut.ep_assert(isinstance(preserved, bool), _("preserved should be bool"))

    if conditions is None:
        conditions = []
    if actions is None:
        actions = []

    conditions = ut.FlattenList(conditions)
    actions = ut.FlattenList(actions)
    tstart = c.NextTrigger()

    # Normal
    if len(conditions) <= 16 and len(actions) <= 64:
        patched_conds = []
        for cond in conditions:
            patched_conds.append(patch_condition(cond))

        patched_actions = []
        for act in actions:
            patched_actions.append(patch_action(act))

        tend = c.RawTrigger(
            conditions=patched_conds,
            actions=patched_actions,
            preserved=preserved,
        )

    else:
        # Extended trigger
        condts = []
        cend = c.Forward()

        # Check conditions
        for i in range(0, len(conditions), 16):
            conds = conditions[i : i + 16]
            cts = c.Forward()

            patched_conds = []
            for cond in conds:
                patched_conds.append(patch_condition(cond))

            nextcond = c.Forward()
            cts << c.RawTrigger(
                nextptr=cend,
                conditions=patched_conds,
                actions=c.SetNextPtr(cts, nextcond),
            )
            nextcond << c.NextTrigger()

            condts.append(cts)

        skipt = c.Forward()
        if not preserved:
            actions.append(c.SetNextPtr(tstart, skipt))

        # Execute actions
        for i in range(0, len(actions), 64):
            acts = actions[i : i + 64]
            patched_actions = []
            for act in acts:
                patched_actions.append(patch_action(act))

            tend = c.RawTrigger(actions=patched_actions)

        if not preserved:
            c.SetNextTrigger(skipt)

        # Revert conditions
        cend << c.NextTrigger()
        for i in range(0, len(condts), 64):
            tend = c.RawTrigger(
                actions=[c.SetNextPtr(cts, cend) for cts in condts[i : i + 64]]
            )

        if not preserved:
            skipt << c.NextTrigger()

    return tstart, tend



## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) import .test_eps_misc;
# (Line 3) import py_warnings;
try:
    test_eps_misc = _RELIMP(".", "test_eps_misc")
except ImportError:
    from . import test_eps_misc
import warnings
# (Line 4) import .report.C240903;
# (Line 5) import .report.C240930;
try:
    C240903 = _RELIMP(".report", "C240903")
except ImportError:
    from .report import C240903
# (Line 7) function square(x) : None;
try:
    C240930 = _RELIMP(".report", "C240930")
except ImportError:
    from .report import C240930
# (Line 9) const a = [
# (Line 10) square(1),
# (Line 11) square(2),
# (Line 12) square(3),
# (Line 13) square(4),
# (Line 14) square(5)
# (Line 15) ];
a = _CGFW(lambda: [_ARR(FlattenList([f_square(1), f_square(2), f_square(3), f_square(4), f_square(5)]))], 1)[0]
# (Line 17) function testLineno() {
@EUDFunc
def f_testLineno():
    # (Line 18) const foo = py_eval("warnings.warn");
    foo = eval("warnings.warn")
    # (Line 19) foo("ㅇㅅㅇ");
    foo("ㅇㅅㅇ")
    # (Line 20) }
    # (Line 22) function square(x) {

@EUDFunc
def f_square(x):
    # (Line 23) testLineno();
    f_testLineno()
    # (Line 24) return x * x; // + z.k;
    EUDReturn(x * x)
    # (Line 25) }
    # (Line 27) const receives = py_eval('[PVariable() for _ in range(8)]');

receives = _CGFW(lambda: [eval('[PVariable() for _ in range(8)]')], 1)[0]
# (Line 28) const attack_gwpID = 4;
attack_gwpID = _CGFW(lambda: [4], 1)[0]
# (Line 29) function constv_thing() {
@EUDFunc
def f_constv_thing():
    # (Line 30) foreach(i, pvar: py_enumerate(receives)) {}
    for i, pvar in enumerate(receives):
        # (Line 31) SetMemoryXEPD(EPD(0x656FB8) + attack_gwpID/4, Add, 100 << (attack_gwpID%4 * 8), 0xFF << (attack_gwpID%4 * 8));  // cooldown +100
        pass

    # (Line 32) return a[0] + a[1] + a[2] + a[3] + a[4];
    DoActions(SetMemoryXEPD(EPD(0x656FB8) + attack_gwpID // 4, Add, _LSH(100,(attack_gwpID % 4 * 8)), _LSH(0xFF,(attack_gwpID % 4 * 8))))
    EUDReturn(a[0] + a[1] + a[2] + a[3] + a[4])
    # (Line 33) }
    # (Line 35) function test_reported() {

@EUDFunc
def f_test_reported():
    # (Line 36) return C240903.updateUnitNameAndRank();
    EUDReturn(C240903.f_updateUnitNameAndRank())
    # (Line 37) }



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Callable
from typing import Any

from .. import core as c
from .. import trigger as tg
from .. import utils as ut
from ..localize import _


def DoActions(*actions, preserved=True) -> tuple[c.Forward, c.RawTrigger]:  # noqa: N802
    return tg.Trigger(actions=actions, preserved=preserved)


def EUDJump(nextptr, *, must_use=True) -> None:  # noqa: N802
    nt_list = ut.EUDGetLastBlockOfName("triggerscope")[1]["nexttrigger_list"]
    if must_use and not nt_list:
        raise ut.EPError(_("unreachable EUDJump"))
    if c.IsEUDVariable(nextptr):
        c.RawTrigger(
            nextptr=nextptr.GetVTable(),
            actions=nextptr.QueueAssignTo(ut.EPD(nextptr.GetVTable()) + 1),
        )
    else:
        c.SetNextTrigger(nextptr)


def EUDJumpIf(conditions, ontrue) -> None:  # noqa: N802
    onfalse = c.Forward()
    tg.EUDBranch(conditions, ontrue, onfalse)
    onfalse << c.NextTrigger()


def EUDJumpIfNot(conditions, onfalse) -> None:  # noqa: N802
    ontrue = c.Forward()
    tg.EUDBranch(conditions, ontrue, onfalse)
    ontrue << c.NextTrigger()


def EUDTernary(  # noqa: N802
    conditions, *, neg=False
) -> Callable[[Any], Callable[[Any], c.EUDVariable]]:
    v = c.EUDVariable()
    t = c.Forward()
    end = c.Forward()

    if neg:
        EUDJumpIf(conditions, t)
    else:
        EUDJumpIfNot(conditions, t)

    def _1(ontrue):
        v << ontrue
        c.SetNextTrigger(end)
        t << c.NextTrigger()

        def _2(onfalse):
            v << onfalse
            end << c.NextTrigger()
            return v

        return _2

    return _1



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from abc import ABCMeta
from typing import TYPE_CHECKING, TypeAlias, TypeVar

from typing_extensions import Self

from ...localize import _
from ...utils import EPError, ExprProxy

if TYPE_CHECKING:
    from ..allocator import ConstExpr
    from ..variable import EUDVariable

Dword: TypeAlias = (
    "int | EUDVariable | ConstExpr | ExprProxy[int | EUDVariable | ConstExpr]"
)
Word: TypeAlias = "int | EUDVariable | ExprProxy[int | EUDVariable]"
Byte: TypeAlias = "int | EUDVariable | ExprProxy[int | EUDVariable]"


class ConstType(ExprProxy, metaclass=ABCMeta):
    __slots__ = ()

    @classmethod
    def cast(cls, _from):
        if isinstance(_from, cls):
            return _from
        if isinstance(_from, ConstType):
            raise EPError(_('"{}" is not a {}').format(_from, cls.__name__))
        return cls(_from)

    def _check_assign(self, other) -> None:
        from ..variable import EUDVariable

        if not isinstance(self._value, EUDVariable):
            raise EPError(_("Can't assign {} to constant expression").format(other))
        if type(other) is ExprProxy:
            other = other._value
        if isinstance(other, type(self)):
            return
        if isinstance(other, int | EUDVariable | str):
            return
        else:
            raise EPError(_("Can't assign {} to {}").format(other, self))

    def Assign(self, other) -> Self:  # noqa: N802
        self._check_assign(other)
        other = type(self).cast(other)
        self._value.Assign(other)
        return self

    def __iadd__(self, other) -> Self:
        self._check_assign(other)
        self._value.__iadd__(other)
        return self

    def __isub__(self, other) -> Self:
        self._check_assign(other)
        self._value.__isub__(other)
        return self

    def __ior__(self, other) -> Self:
        self._check_assign(other)
        self._value.__ior__(other)
        return self

    def __iand__(self, other) -> Self:
        self._check_assign(other)
        self._value.__iand__(other)
        return self

    def __ixor__(self, other) -> Self:
        self._check_assign(other)
        self._value.__ixor__(other)
        return self

    def __imul__(self, other) -> Self:
        self._check_assign(other)
        self._value.__imul__(other)
        return self

    def __ifloordiv__(self, other) -> Self:
        self._check_assign(other)
        self._value.__ifloordiv__(other)
        return self

    def __imod__(self, other) -> Self:
        self._check_assign(other)
        self._value.__imod__(other)
        return self

    def __ilshift__(self, other) -> Self:
        self._check_assign(other)
        self._value.__ilshift__(other)
        return self

    def __irshift__(self, other) -> Self:
        self._check_assign(other)
        self._value.__irshift__(other)
        return self


# return types
_Dword: TypeAlias = "int | EUDVariable | ConstExpr"
_Word: TypeAlias = "int | EUDVariable"
_Byte: TypeAlias = "int | EUDVariable"

# argument types
T = TypeVar("T", int, "EUDVariable", "ConstExpr")
U = TypeVar("U", int, "EUDVariable")
_ExprProxy: TypeAlias = (
    "ExprProxy[ConstType | int | EUDVariable | ConstExpr | _ExprProxy]"
)
_Arg: TypeAlias = "ConstType | int | EUDVariable | ConstExpr | ExprProxy[_Arg]"
__ExprProxy: TypeAlias = "ExprProxy[ConstType | int | EUDVariable | __ExprProxy]"
__Arg: TypeAlias = "ConstType | int | EUDVariable | ExprProxy[__Arg]"



from helper import *


@TestInstance
def test_variable_hptr():
    a = EUDVariable(5)
    b = a * a * a * a * a

    # variable test
    x = Db(1024)
    y = Db(1024)
    f_dbstr_print(x, "    \x04test ", a, " b: ", hptr(b), " test", 1, hptr(21))
    f_dbstr_print(y, "    \x04test 5 b: 00000C35 test100000015")
    test_equality("variable/hptr printing test", f_strcmp(x, y), 0)


@TestInstance
def test_strprint():
    x = Db(1024)
    y = Db("Test instance")
    f_dbstr_print(x, ptr2s(y))
    f_printAll("{:s}", x)
    f_printAll("{:s}", y)
    test_equality("string printing test", f_strcmp(x, y), 0)



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from math import ceil

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..collections.eudarray import EUDArray
from ..collections.playerv import PVariable
from ..localize import _
from ..memio import (
    f_dwbreak,
    f_dwread_epd,
    f_getcurpl,
    f_memcmp,
    f_setcurpl,
    f_wread_epd,
)
from ..scdata import CurrentPlayer, TrgPlayer
from ..trigger import Trigger
from .cpprint import f_cpstr_print
from .eudprint import f_dbstr_print
from .fmtprint import _format_args
from .strcommon import epd2s, ptr2s
from .strfunc import f_strlen_epd

PLVarUnit, PLVarMask = ceil((0x58F500 - 0x58A364) / 48), 0
# FIXME: PLVarDict = {}


def _get_player_lightvar() -> tuple[int, int]:
    global PLVarUnit, PLVarMask
    ret = (PLVarUnit, 1 << PLVarMask)
    PLVarMask += 1
    if PLVarMask >= 32:
        PLVarMask = 0
        PLVarUnit += 1
    return ret


def compare_sequence(src: int, seq: str) -> list[c.Condition]:
    ut.ep_assert(isinstance(src, int), f"Invalid address: {src}")
    ut.ep_assert(isinstance(seq, str), f"Invalid string: {seq}")
    seq += "\0"
    ret = []
    val, mask = 0, 0

    def append_cmp() -> None:
        nonlocal val, mask
        ret.append(c.MemoryX(src - src % 4, c.Exactly, val, mask))
        val, mask = 0, 0

    for char in seq:
        lsh = 8 * (src % 4)
        val += ord(char) << lsh
        mask += 0xFF << lsh
        if src % 4 == 3:
            append_cmp()
        src += 1
    if mask:
        append_cmp()
    return ret


def _get_ispname_cond(name, _plvars={}):
    try:
        init, end, params = _plvars[name]
    except KeyError:
        c.PushTriggerScope()
        unit, mask = _get_player_lightvar()
        init, params = c.NextTrigger(), (c.AtLeast, 1, unit, mask)
        end = c.Forward()
        _plvars[name] = (init, end, params)
        if isinstance(name, str):
            # TODO: only iterate human players
            for player in range(8):
                t = c.RawTrigger(
                    conditions=compare_sequence(0x57EEEB + 36 * player, name),
                    actions=c.SetDeaths(player, c.Add, mask, unit),
                    preserved=False,
                )
        else:
            # TODO: compare every names in a single loop
            from ..eudlib.utilf.pexist import EUDLoopPlayer

            for player in EUDLoopPlayer(None):
                equals = f_memcmp(0x57EEEB + 36 * player, name, 25)
                Trigger(
                    conditions=equals.Exactly(0),
                    actions=c.SetDeaths(player, c.Add, mask, unit),
                    preserved=False,
                )
            t = c.RawTrigger()
        end << t
        c.PopTriggerScope()
    return init, end, params


def IsPName(player, name):  # noqa: N802
    p = c.EncodePlayer(player)
    if isinstance(p, int):
        if p < 8:
            if isinstance(name, str):
                return compare_sequence(0x57EEEB + 36 * p, name)
        else:
            # fmt: off
            ut.ep_assert(
                p == c.EncodePlayer(CurrentPlayer),
                _("IsPName player should be Player1 to Player8 or CurrentPlayer, not {}").format(player),  # noqa: E501
            )
            # fmt: on

    init, end, params = _get_ispname_cond(name)
    if cs.EUDExecuteOnce()():
        # TODO: initialize when game starts
        nptr = c.Forward()
        c.RawTrigger(nextptr=init, actions=c.SetNextPtr(end, nptr))
        nptr << c.NextTrigger()
    cs.EUDEndExecuteOnce()
    return c.DeathsX(p, *params)


class _PlayerName:
    def __init__(self):
        self.is_chatptr_unchanged = c.EUDLightBool()
        self.db = c.Db(218)
        self.ptr, self.epd = c.EUDCreateVariables(2)
        self.odds_or_even = c.EUDLightBool()
        self.name_db_list = [c.Db(26) for _ in range(8)]
        self.basenames, self.baselens = PVariable(self.name_db_list), PVariable()
        self.optimize_start, self.optimize_end = c.Forward(), c.Forward()
        self.odd = [c.Forward() for p in range(8)]
        self.even0 = [c.Forward() for p in range(8)]
        self.even1 = [c.Forward() for p in range(8)]

    def _optimize(self):
        from ..eudlib.utilf.gametick import f_getgametick
        from ..eudlib.utilf.pexist import EUDLoopPlayer

        if self.optimize_start.IsSet():
            return

        c.PushTriggerScope()
        prev_textptr, _end = c.Forward(), c.Forward()
        self.optimize_start << c.NextTrigger()

        odd_arr = EUDArray([ut.EPD(x) for x in self.odd])
        even0_arr = EUDArray([ut.EPD(x) for x in self.even0])
        even1_arr = EUDArray([ut.EPD(x) for x in self.even1])

        if cs.EUDExecuteOnce()():
            for player in EUDLoopPlayer(None):
                player_name = 36 * player
                playerid_epd = 9 * player
                cs.DoActions(
                    player_name.AddNumber(0x57EEEB),
                    playerid_epd.AddNumber(ut.EPD(0x57EEEC)),
                )
                idlen = f_strlen_epd(playerid_epd)
                name_db = self.basenames[player]
                f_dbstr_print(name_db, ptr2s(player_name), ":")
                val_odd = f_dwread_epd(ut.EPD(name_db))
                v0, v1 = f_dwbreak(val_odd)[0:2]
                val_even0 = v0 * 0x10000
                val_even1 = v1 + f_wread_epd(ut.EPD(name_db) + 1, 0) * 0x10000
                con_odd, con_even = odd_arr[player], even1_arr[player]
                cs.DoActions(
                    idlen.AddNumber(1),  # fix ":" not recognized
                    player_name.AddNumber(1),
                    c.SetMemoryEPD(con_odd + 2, c.SetTo, val_odd),
                    c.SetMemoryEPD(even0_arr[player] + 2, c.SetTo, val_even0),
                    c.SetMemoryEPD(con_even + 2, c.SetTo, val_even1),
                )
                f_dbstr_print(name_db, ptr2s(player_name), ":")
                self.baselens[player] = idlen
                if cs.EUDIf()(idlen <= 4):
                    cs.DoActions(
                        c.SetMemoryEPD(con_even + 2, c.SetTo, 0),
                        c.SetMemoryEPD(con_even + 3, c.SetTo, 0x0F000000),
                    )
                    if cs.EUDIf()(idlen <= 2):
                        cs.DoActions(
                            c.SetMemoryEPD(con_odd + 2, c.SetTo, 0),
                            c.SetMemoryEPD(con_odd + 3, c.SetTo, 0x0F000000),
                        )
                    cs.EUDEndIf()
                cs.EUDEndIf()
        cs.EUDEndExecuteOnce()

        once = c.Forward()
        cs.EUDJumpIf([once << c.Memory(0x57F23C, c.Exactly, ~0)], self.optimize_end)
        cs.DoActions(self.is_chatptr_unchanged.Clear())
        f_getgametick(ret=[ut.EPD(once) + 2])
        c.RawTrigger(
            conditions=[prev_textptr << c.Memory(0x640B58, c.Exactly, 11)],
            actions=self.is_chatptr_unchanged.Set(),
        )
        cs.EUDJumpIf(self.is_chatptr_unchanged, _end)
        cs.DoActions(c.SetMemory(prev_textptr + 8, c.SetTo, 0))
        for i in range(3, -1, -1):
            c.RawTrigger(
                conditions=c.MemoryX(0x640B58, c.AtLeast, 1, 2**i),
                actions=c.SetMemory(prev_textptr + 8, c.Add, 2**i),
            )
        _end << c.NextTrigger()
        self.optimize_end << c.RawTrigger()
        c.PopTriggerScope()

    def check_id(self, player, dst):
        ret = c.EUDVariable()
        ret << 1

        cs.EUDSwitch(player)
        for i in range(8):
            cs.EUDSwitchCase()(i)
            if cs.EUDIf()(
                [
                    self.odds_or_even.IsCleared(),
                    self.odd[i] << c.MemoryEPD(dst, c.Exactly, 0),
                ]
            ):
                ret << 0

            if cs.EUDElseIf()(
                [
                    self.odds_or_even,
                    self.even0[i] << c.MemoryXEPD(dst, c.Exactly, 0, 0xFFFF0000),
                    self.even1[i] << c.MemoryEPD(dst + 1, c.Exactly, 0),
                ]
            ):
                ret << 0
            cs.EUDEndIf()
            cs.EUDBreak()
        cs.EUDEndSwitch()

        return ret

    def set_pname(self, player, *name):
        from ..eudlib.utilf.pexist import f_playerexist

        self._optimize()

        start = c.Forward()
        c.RawTrigger(
            nextptr=self.optimize_start,
            actions=c.SetNextPtr(self.optimize_end, start),
        )
        start << c.NextTrigger()

        _end = c.Forward()
        cs.EUDJumpIf(self.is_chatptr_unchanged, _end)
        if isinstance(player, TrgPlayer):
            player = c.EncodePlayer(player)
            if not c.IsEUDVariable(player) and player == CurrentPlayer:
                player = f_getcurpl()
        cs.EUDJumpIf(f_playerexist(player) == 0, _end)
        basename, baselen = self.basenames[player], self.baselens[player]

        origcp = f_getcurpl()
        cs.DoActions(
            self.ptr.SetNumber(0x640B61),
            self.epd.SetNumber(ut.EPD(0x640B60)),
            self.odds_or_even.Clear(),
        )
        if cs.EUDWhile()(self.ptr.AtMost(0x640B61 + 218 * 10)):
            cs.EUDContinueIf(f_check_id(player, self.epd))
            cs.EUDContinueIf(f_memcmp(basename, self.ptr, baselen) >= 1)
            cs.DoActions(
                c.SetCurrentPlayer(ut.EPD(self.db)), baselen.SubtractNumber(1)
            )  # fix setpname save 1byte less chat contents
            f_cpstr_print(ptr2s(self.ptr + baselen), EOS=False)
            cs.DoActions(
                c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0),
                baselen.AddNumber(1),
                c.SetCurrentPlayer(self.epd),
            )
            c.RawTrigger(
                conditions=self.odds_or_even.IsSet(),
                actions=[
                    c.SetDeaths(CurrentPlayer, c.SetTo, ut.b2i4(b"\r" * 4), 0),
                    c.AddCurrentPlayer(1),
                ],
            )
            name = list(name)
            name.append(epd2s(ut.EPD(self.db)))
            f_cpstr_print(*name)

            cs.EUDSetContinuePoint()
            cs.DoActions(
                self.ptr.AddNumber(218),
                self.epd.AddNumber(54),
                self.odds_or_even.Toggle(),
            )
            c.RawTrigger(
                conditions=self.odds_or_even.IsCleared(),
                actions=self.epd.AddNumber(1),
            )
        cs.EUDEndWhile()
        f_setcurpl(origcp)

        _end << c.NextTrigger()


_global_pname_class = _PlayerName()


@c.EUDTypedFunc([TrgPlayer, None])
def f_check_id(player, dst):
    return _global_pname_class.check_id(player, dst)


def SetPName(player, *name):  # noqa: N802
    _global_pname_class.set_pname(player, *name)


def SetPNamef(player, format_string, *args):  # noqa: N802
    fmtargs = _format_args(format_string, *args)
    SetPName(player, *fmtargs)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import os
import sys

from .... import utils as ut
from ....localize import _
from ...allocator import Forward
from ...eudobj import Db
from ...rawtrigger import (
    Add,
    AtLeast,
    MemoryX,
    NextTrigger,
    PopTriggerScope,
    PushTriggerScope,
    RawTrigger,
    SetMemory,
    SetMemoryEPD,
    SetNextPtr,
    SetTo,
    Subtract,
)
from ...variable import EUDVariable, SeqCompute
from . import tracecrypt

i_header: bytes = os.urandom(16)
_tracetool_data_epd = ut.EPD(Db(i_header + bytes(4 * 2048)))
_record_trace_act = Forward()
PushTriggerScope()
_record_trace_trigger = RawTrigger(
    actions=[_record_trace_act << SetMemoryEPD(_tracetool_data_epd + 4, SetTo, 0)]
)
PopTriggerScope()


def _f_initstacktrace() -> None:
    # Fill the header of trace stack with random string on runtime.
    # SC:R creates a copy of STR section during EUD emulation and writes to
    # only one of the copy during emulation stage. We should read the copy
    # that is really written in-game.
    # We will fill the trace stack 'magic code' on runtime, and later find the
    # magic code to locate the stack trace table.
    if not trace_header:
        raise ut.EPError(_("Must call SaveMap first"))
    RawTrigger(
        actions=[
            SetMemoryEPD(_tracetool_data_epd + 0, SetTo, ut.b2i4(trace_header, 0x0)),
            SetMemoryEPD(_tracetool_data_epd + 1, SetTo, ut.b2i4(trace_header, 0x4)),
            SetMemoryEPD(_tracetool_data_epd + 2, SetTo, ut.b2i4(trace_header, 0x8)),
            SetMemoryEPD(_tracetool_data_epd + 3, SetTo, ut.b2i4(trace_header, 0xC)),
        ],
    )


def _eud_trace_push() -> None:
    RawTrigger(actions=SetMemory(_record_trace_act + 16, Add, 1))


def _eud_trace_pop() -> None:
    _eud_trace_log_raw(0)
    RawTrigger(actions=SetMemory(_record_trace_act + 16, Subtract, 1))


_next_trace_id = 0
trace_map: list[tuple[int, str]] = []
_trace_key = 0
trace_header: bytes = b""


def GetTraceStackDepth() -> EUDVariable:  # noqa: N802
    v = EUDVariable()
    v << 0
    for i in range(31, -1, -1):
        RawTrigger(
            conditions=MemoryX(_record_trace_act + 16, AtLeast, 1, 2**i),
            actions=v.AddNumber(2**i),
        )
    SeqCompute([(v, Subtract, _tracetool_data_epd + 4)])
    return v


def _reset_trace_map() -> None:
    """This function gets called by savemap.py::SaveMap to clear trace data."""
    global _next_trace_id, _trace_key, trace_header
    _next_trace_id = 0
    _trace_key = ut.b2i4(os.urandom(4))
    trace_map.clear()
    trace_header = os.urandom(16)


def EUDTraceLog(lineno: int | None = None) -> None:  # noqa: N802
    """Log trace."""
    global _next_trace_id

    # Construct trace message from cpython stack
    # Note: we need to get the caller's filename, function name, and line no.
    # Using inspect module for this purpose is insanely slow, so we use
    # plain sys object with plain frame attributes.

    frame = sys._getframe(1)
    try:
        if lineno is None:
            lineno = frame.f_lineno
        msg = f"{frame.f_code.co_filename}|{frame.f_code.co_name}|{lineno}"
    finally:
        # frame object should be dereferenced as quickly as possible.
        # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
        del frame

    v = tracecrypt.mix(_trace_key, _next_trace_id)
    _next_trace_id += 1
    if v == 0:  # We don't allow logging 0.
        v = tracecrypt.mix(_trace_key, _next_trace_id)
        _next_trace_id += 1
    trace_map.append((v, str(msg)))

    _eud_trace_log_raw(v)


def _eud_trace_log_raw(v: int) -> None:
    nt = Forward()
    RawTrigger(
        nextptr=_record_trace_trigger,
        actions=[
            SetNextPtr(_record_trace_trigger, nt),
            SetMemory(_record_trace_act + 20, SetTo, v),
        ],
    )
    nt << NextTrigger()


def _get_trace_map() -> tuple[tuple[bytes, bytes], list[tuple[int, str]]]:
    return (i_header, trace_header), trace_map



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .basicstru import DoActions, EUDJump, EUDJumpIf, EUDJumpIfNot, EUDTernary
from .breakcont import (
    EUDBreak,
    EUDBreakIf,
    EUDBreakIfNot,
    EUDContinue,
    EUDContinueIf,
    EUDContinueIfNot,
    EUDIsContinuePointSet,
    EUDSetContinuePoint,
)
from .cshelper import CtrlStruOpener
from .logic import EUDAnd, EUDNot, EUDOr
from .loopblock import (
    EUDEndInfLoop,
    EUDEndLoopN,
    EUDEndWhile,
    EUDInfLoop,
    EUDLoopN,
    EUDLoopRange,
    EUDWhile,
    EUDWhileNot,
)
from .shortcircuit import EUDSCAnd, EUDSCOr
from .simpleblock import (
    EUDElse,
    EUDElseIf,
    EUDElseIfNot,
    EUDEndExecuteOnce,
    EUDEndIf,
    EUDExecuteOnce,
    EUDIf,
    EUDIfNot,
)
from .swblock import (
    EPDSwitch,
    EUDEndSwitch,
    EUDSwitch,
    EUDSwitchCase,
    EUDSwitchDefault,
)

__all__ = [
    "DoActions",
    "EUDJump",
    "EUDJumpIf",
    "EUDJumpIfNot",
    "EUDTernary",
    "EUDBreak",
    "EUDBreakIf",
    "EUDBreakIfNot",
    "EUDContinue",
    "EUDContinueIf",
    "EUDContinueIfNot",
    "EUDIsContinuePointSet",
    "EUDSetContinuePoint",
    "CtrlStruOpener",
    "EUDAnd",
    "EUDNot",
    "EUDOr",
    "EUDEndInfLoop",
    "EUDEndLoopN",
    "EUDEndWhile",
    "EUDInfLoop",
    "EUDLoopN",
    "EUDLoopRange",
    "EUDWhile",
    "EUDWhileNot",
    "EUDSCAnd",
    "EUDSCOr",
    "EUDElse",
    "EUDElseIf",
    "EUDElseIfNot",
    "EUDEndExecuteOnce",
    "EUDEndIf",
    "EUDExecuteOnce",
    "EUDIf",
    "EUDIfNot",
    "EPDSwitch",
    "EUDEndSwitch",
    "EUDSwitch",
    "EUDSwitchCase",
    "EUDSwitchDefault",
]



# Copyright 2022 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ..utils import EPD, ep_assert
from .rawtrigger import (
    Add,
    RawTrigger,
    SetDeaths,
    SetMemory,
    SetMemoryEPD,
    SetMemoryX,
    SetMemoryXEPD,
    SetTo,
    Subtract,
)
from .variable import IsEUDVariable, VProc


def cpset(a, b):
    if not (IsEUDVariable(a) or IsEUDVariable(b)):
        return a + b, RawTrigger
    elif IsEUDVariable(a) and IsEUDVariable(b):
        if a is b:
            VProc(a, [SetMemory(0x6509B0, SetTo, 0), *a.QueueAddTo(EPD(0x6509B0))])
            VProc(a, [])
        else:
            VProc(
                [a, b],
                [
                    *a.QueueAssignTo(EPD(0x6509B0)),
                    *b.QueueAddTo(EPD(0x6509B0)),
                ],
            )
    else:
        if IsEUDVariable(b):
            a, b = b, a
        VProc(
            a,
            [
                SetMemory(0x6509B0, SetTo, b),
                *a.QueueAddTo(EPD(0x6509B0)),
            ],
        )
    from ..memio.modcurpl import f_setcurpl2cpcache

    return 13, f_setcurpl2cpcache  # CurrentPlayer


def iset(a, b, modifier, v):
    """SetMemoryEPD(a + b, modifier, v)"""
    if not (IsEUDVariable(a) or IsEUDVariable(b)) or a is b or a is v or b is v:
        from ..memio.dwepdio import setdw_epd

        return setdw_epd(a + b, modifier, v)
    if IsEUDVariable(v):
        if IsEUDVariable(a) and IsEUDVariable(b):
            return VProc(
                [a, b, v],
                [
                    *a.QueueAssignTo(EPD(v.getDestAddr())),
                    *b.QueueAddTo(EPD(v.getDestAddr())),
                    v.SetModifier(modifier),
                ],
            )
        if IsEUDVariable(b):
            a, b = b, a
        return VProc(
            [a, v],
            [
                v.SetDest(b),
                *a.QueueAddTo(EPD(v.getDestAddr())),
                v.SetModifier(modifier),
            ],
        )
    set_v = SetDeaths(0, modifier, v, 0)
    if IsEUDVariable(a) and IsEUDVariable(b):
        VProc(
            [a, b],
            [
                *a.QueueAssignTo(EPD(set_v) + 4),
                *b.QueueAddTo(EPD(set_v) + 4),
            ],
        )
    else:
        if IsEUDVariable(b):
            a, b = b, a
        VProc(
            a,
            [
                SetMemory(set_v + 16, SetTo, b),
                *a.QueueAddTo(EPD(set_v) + 4),
            ],
        )
    return RawTrigger(actions=set_v)


def isub(a, b, v):
    if not IsEUDVariable(v) or a is b or a is v or b is v:
        return iset(a, b, Add, -v)
    dst, trg = cpset(a, b)
    VProc(
        v,
        [
            SetMemoryXEPD(dst, Add, -1, 0x55555555),
            SetMemoryXEPD(dst, Add, -1, 0xAAAAAAAA),
            SetMemoryEPD(dst, Add, 1),
            *v.QueueAddTo(dst),
        ],
    )
    return trg(
        actions=[
            SetMemoryXEPD(dst, Add, -1, 0x55555555),
            SetMemoryXEPD(dst, Add, -1, 0xAAAAAAAA),
            SetMemoryEPD(dst, Add, 1),
        ],
    )


def iand(a, b, v):
    if not IsEUDVariable(v):
        if not (IsEUDVariable(a) or IsEUDVariable(b)):
            return RawTrigger(actions=SetMemoryXEPD(a + b, SetTo, 0, ~v))

        write = SetMemoryXEPD(0, SetTo, 0, ~v)
        if IsEUDVariable(a) and IsEUDVariable(b):
            VProc(
                [a, b],
                [
                    *a.QueueAssignTo(EPD(write) + 4),
                    *b.QueueAddTo(EPD(write) + 4),
                ],
            )
        else:
            if IsEUDVariable(b):
                a, b = b, a
            VProc(
                a,
                [
                    SetMemory(write + 16, SetTo, b),
                    *a.QueueAddTo(EPD(write) + 4),
                ],
            )
        return RawTrigger(actions=write)

    if not (IsEUDVariable(a) or IsEUDVariable(b)):
        write = SetMemoryXEPD(a + b, SetTo, 0, 0)
        VProc(v, v.QueueAssignTo(EPD(write)))
    elif IsEUDVariable(a) and IsEUDVariable(b):
        write = SetMemoryXEPD(0, SetTo, 0, 0)
        VProc(
            [a, b, v],
            [
                *a.QueueAssignTo(EPD(write) + 4),
                *b.QueueAddTo(EPD(write) + 4),
                *v.QueueAssignTo(EPD(write)),
            ],
        )
    else:
        write = SetMemoryXEPD(0, SetTo, 0, 0)
        if IsEUDVariable(b):
            a, b = b, a
        VProc(
            [a, v],
            [
                SetMemory(write + 16, SetTo, b),
                *a.QueueAddTo(EPD(write) + 4),
                *v.QueueAssignTo(EPD(write)),
            ],
        )
    return RawTrigger(
        actions=[
            SetMemoryX(write, Add, -1, 0x55555555),
            SetMemoryX(write, Add, -1, 0xAAAAAAAA),
            write,
        ],
    )


def ior(a, b, v):
    if not IsEUDVariable(v):
        if not (IsEUDVariable(a) or IsEUDVariable(b)):
            return RawTrigger(actions=SetMemoryXEPD(a + b, SetTo, ~0, v))

        write = SetMemoryXEPD(0, SetTo, ~0, v)
        if IsEUDVariable(a) and IsEUDVariable(b):
            VProc(
                [a, b],
                [
                    *a.QueueAssignTo(EPD(write) + 4),
                    *b.QueueAddTo(EPD(write) + 4),
                ],
            )
        else:
            if IsEUDVariable(b):
                a, b = b, a
            VProc(
                a,
                [
                    SetMemory(write + 16, SetTo, b),
                    *a.QueueAddTo(EPD(write) + 4),
                ],
            )
        return RawTrigger(actions=write)

    if not (IsEUDVariable(a) or IsEUDVariable(b)):
        write = SetMemoryXEPD(a + b, SetTo, ~0, 0)
        VProc(v, v.QueueAssignTo(EPD(write)))
        return RawTrigger(actions=write)

    write = SetMemoryXEPD(0, SetTo, ~0, 0)
    if IsEUDVariable(a) and IsEUDVariable(b):
        VProc(
            [a, b, v],
            [
                *a.QueueAssignTo(EPD(write) + 4),
                *b.QueueAddTo(EPD(write) + 4),
                *v.QueueAssignTo(EPD(write)),
            ],
        )
    else:
        if IsEUDVariable(b):
            a, b = b, a
        VProc(
            [a, v],
            [
                SetMemory(write + 16, SetTo, b),
                *a.QueueAddTo(EPD(write) + 4),
                *v.QueueAssignTo(EPD(write)),
            ],
        )
    return RawTrigger(actions=write)


def ixor(a, b, v):
    if not IsEUDVariable(v):
        dst, trg = cpset(a, b)
        return trg(
            actions=[
                SetMemoryXEPD(dst, Add, v, 0x55555555),
                SetMemoryXEPD(dst, Add, v, 0xAAAAAAAA),
            ],
        )

    dst = EPD(v.getDestAddr())
    if not (IsEUDVariable(a) or IsEUDVariable(b)):
        VProc(v, [*v.QueueAddTo(a + b), v.SetMask(0x55555555)])
    elif IsEUDVariable(a) and IsEUDVariable(b):
        VProc(
            [a, b, v],
            [
                *a.QueueAssignTo(dst),
                *b.QueueAddTo(dst),
                v.SetMask(0x55555555),
                v.SetModifier(Add),
            ],
        )
    else:
        if IsEUDVariable(b):
            a, b = b, a
        VProc(
            [a, v],
            [
                *a.QueueAddTo(dst),
                *v.QueueAddTo(b),
                v.SetMask(0x55555555),
            ],
        )
    VProc(v, v.SetMask(0xAAAAAAAA))
    # FIXME: restore to previous mask???
    return RawTrigger(actions=v.SetMask(0xFFFFFFFF))


def ilshift(a, b, n):
    ep_assert(isinstance(n, int), "left shift amount should be constant")
    if n == 0:
        return
    mask = (1 << (n + 1)) - 1
    dst, trg = cpset(a, b)

    def itemw(mod, value, mask):
        return SetMemoryXEPD(dst, mod, value, mask)

    return trg(
        actions=[
            [
                itemw(SetTo, 0, (mask >> 1) << (n + 1)),
                itemw(Add, (mask >> 1) << n, mask << n),
            ]
            for n in reversed(range(32 - n))
        ]
        + [itemw(SetTo, 0, mask >> 1)]  # lowest n bits
    )


def irshift(a, b, n):
    ep_assert(isinstance(n, int), "right shift amount should be constant")
    if n == 0:
        return
    mask = (1 << (n + 1)) - 1
    dst, trg = cpset(a, b)

    def sub(value, mask):
        return SetMemoryXEPD(dst, Subtract, value, mask)

    return trg(
        actions=[SetMemoryXEPD(dst, SetTo, 0, mask >> 1)]  # lowest n bits
        + [sub((mask >> 1) << n, mask << n) for n in range(32 - n)]
    )



import os
import sys
import time

_currentProfileTree = None


_last_time = time.process_time()


def _profiler(frame, event, arg):
    global _last_time
    global _currentProfileTree

    currentTime = time.process_time()
    elapsedTime = currentTime - _last_time

    if not _currentProfileTree:
        return

    if event == "call":
        try:
            newTree = _currentProfileTree[frame.f_code]
            newTree["_cumtime2"] = 0
            newTree["_calln"] += 1
        except KeyError:
            newTree = {
                "_code": frame.f_code,
                "_parent": _currentProfileTree,
                "_time": 0,
                "_calln": 1,
                "_cumtime": 0,
                "_cumtime2": 0,
            }
            _currentProfileTree[frame.f_code] = newTree
        _currentProfileTree["_time"] += elapsedTime
        _currentProfileTree["_cumtime"] += elapsedTime
        _currentProfileTree["_cumtime2"] += elapsedTime
        _currentProfileTree = newTree

        _last_time = time.process_time()

    elif event == "return":
        _currentProfileTree["_time"] += elapsedTime
        _currentProfileTree["_cumtime"] += elapsedTime

        parent = _currentProfileTree["_parent"]
        if parent:
            ct2 = _currentProfileTree["_cumtime2"] + elapsedTime
            parent["_cumtime"] += ct2
            parent["_cumtime2"] += ct2
            _currentProfileTree = parent
        else:
            _currentProfileTree = None

        _last_time = time.process_time()


def profile(f, ofname):
    global _currentProfileTree, _last_time
    _currentProfileTree = {
        "_code": None,
        "_parent": None,
        "_time": 0,
        "_cumtime": 0,
        "_cumtime2": 0,
    }
    rootTree = _currentProfileTree
    _last_time = time.process_time()
    sys.setprofile(_profiler)
    f()
    sys.setprofile(None)

    with open(ofname, "w") as output:
        _writeTree(output, rootTree, 0)


def _writeTree(output, tree, indent):
    header = "  " * indent
    for key in sorted(
        filter(lambda x: type(x) is not str, tree.keys()),
        key=lambda x: -tree[x]["_cumtime"],
    ):
        subtree = tree[key]
        code = subtree["_code"]
        if not code.co_name or code.co_name[0] == "_":
            continue

        sourcefname = os.path.split(code.co_filename)[-1]
        output.write(
            '%s"%s@%s:%d(%d / %f,%f)": {\n'
            % (
                header,
                code.co_name,
                sourcefname,
                code.co_firstlineno,
                subtree["_calln"],
                subtree["_cumtime"],
                subtree["_time"],
            )
        )
        _writeTree(output, subtree, indent + 1)
        output.write("%s},\n" % header)



from helper import *


class expect_triggerscope_error:
    def __enter__(self):
        pass

    def __exit__(self, type, e, traceback):
        if isinstance(e, TriggerScopeError):
            print(" - Error as expected : %s" % e)
            return True
        else:
            raise RuntimeError("EPError not thrown")


with expect_triggerscope_error():
    RawTrigger()


@EUDFunc
def f_add(a, b):
    return a + b


@EUDFunc
def f_mul(a, b):
    return a * b


@EUDFunc
def f_div(a, b):
    return a // b, a % b


@EUDFunc
def f_addmul(a, b):
    return a + b, a * b


noretCheck = EUDVariable()


@EUDFunc
def testnoret():
    noretCheck << 1


@TestInstance
def test_dfptr():
    p = EUDFuncPtr(2, 1).alloc()

    # Test - Simple pattern
    p << f_add
    a = p(7, 9)

    p << f_mul
    b = p(7, 9)

    with expect_eperror():
        p << f_div  # Should trigger error

    test_equality("Simple dfptr call", [a, b], [16, 63])

    # Test2 - multi return
    q = EUDFuncPtr(2, 2).alloc()
    q << f_div
    a1, a2 = q(12, 5)
    q << f_addmul
    b1, b2 = q(12, 5)
    q << f_div
    c1, c2 = q(17, 3)
    test_equality(
        "Function dpointer test", [a1, a2, b1, b2, c1, c2], [2, 2, 17, 60, 5, 2]
    )

    # Test 3 - no arg no ret
    r = EUDFuncPtr(0, 0).alloc()
    r << testnoret
    r()
    test_assert("dFunction pointer test (0.0 Func)", noretCheck == 1)

    # Test 4 - transfer fptr to fptr
    r = EUDFuncPtr(2, 1).cast(p)
    p << f_mul
    a, b = p(7, 9), r(7, 9)
    r << f_add
    c, d = p(7, 9), r(7, 9)
    test_equality("Indirect dfptr call", [a, b, c, d], [63, 63, 16, 16])

    # Test 5 - fptr cloning
    s = p.copy()
    p << f_mul
    e, f = p(7, 9), s(7, 9)
    test_equality("dFptr cloning", [e, f], [63, 16])


# Nested fptr call
@EUDTypedFunc([EUDFuncPtr(1, 1)])
def f1(f):
    return 4 + f(4) + 6


@EUDFunc
def f(x):
    return x * x


@TestInstance
def test_nested_dfptr():
    k = EUDFuncPtr(1, 1).alloc()
    k << f1
    test_equality("Nested dfptr", k(EUDFuncPtr(1, 1)(f)), 26)



from helper import *


@EUDTypedFunc([TrgLocation, TrgString])
def x(a, b):
    test_equality(
        "Location and strings can be used as a parameter",
        [a, b],
        [EncodeLocation("Anywhere"), EncodeString("Hello world!")],
    )


@TestInstance
def test_mdhelper():
    assert TrgLocation.cast("Anywhere") == 64
    assert TrgString.cast("Hello world!") == EncodeString("Hello world!")
    assert TrgSwitch.cast("Switch 1") == 0
    assert TrgUnit.cast("Terran Marine") == 0
    x("Anywhere", "Hello world!")



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing_extensions import Self

from ...core.allocator.payload import ObjCollector
from .allocator import ConstExpr, ObjAllocator, PayloadBuffer, RlocInt_C

class EUDObject(ConstExpr):
    def __new__(cls) -> Self: ...
    def DynamicConstructed(self) -> bool: ...  # noqa: N802
    def Evaluate(self) -> RlocInt_C: ...  # noqa: N802
    def GetDataSize(self) -> int: ...  # noqa: N802
    def CollectDependency(self, pbuf: ObjCollector) -> None: ...  # noqa: N802
    def WritePayload(self, pbuf: ObjAllocator | PayloadBuffer) -> None: ...  # noqa: N802



# Copyright 2022 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .buttonset import DefaultButtonSet, DefButtonSetDict
from .flingy import DefaultFlingy, DefFlingyDict
from .icon import DefaultIcon, DefIconDict
from .image import DefaultImage, DefImageDict
from .iscript import DefaultIscript, DefIscriptDict
from .portrait import DefaultPortrait, DefPortraitDict
from .sfxdata import DefaultSfxData, DefSfxDataDict
from .sprite import DefaultSprite, DefSpriteDict
from .stattxt import DefaultRank, DefaultStatText, DefRankDict, DefStatTextDict
from .tech import DefaultTech, DefTechDict
from .trg import (
    DefAIScriptDict,
    DefaultAIScriptAtLocation,
    DefaultAIScriptWithoutLocation,
    DefaultUnit,
    DefLocationDict,
    DefSwitchDict,
    DefUnitDict,
)
from .unitorder import DefaultUnitOrder, DefUnitOrderDict
from .upgrade import DefaultUpgrade, DefUpgradeDict
from .weapon import DefaultWeapon, DefWeaponDict

__all__ = [
    "DefaultButtonSet",
    "DefButtonSetDict",
    "DefaultFlingy",
    "DefFlingyDict",
    "DefaultIcon",
    "DefIconDict",
    "DefaultImage",
    "DefImageDict",
    "DefaultIscript",
    "DefIscriptDict",
    "DefaultPortrait",
    "DefPortraitDict",
    "DefaultSfxData",
    "DefSfxDataDict",
    "DefaultSprite",
    "DefSpriteDict",
    "DefaultRank",
    "DefaultStatText",
    "DefRankDict",
    "DefStatTextDict",
    "DefaultTech",
    "DefTechDict",
    "DefAIScriptDict",
    "DefaultAIScriptAtLocation",
    "DefaultAIScriptWithoutLocation",
    "DefaultUnit",
    "DefLocationDict",
    "DefSwitchDict",
    "DefUnitDict",
    "DefaultUnitOrder",
    "DefUnitOrderDict",
    "DefaultUpgrade",
    "DefUpgradeDict",
    "DefaultWeapon",
    "DefWeaponDict",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...bindings._rust import eudobj

EUDObject = eudobj.EUDObject



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing_extensions import Self

from ...localize import _
from ...utils import EPError, unProxy
from .. import rawtrigger as bt
from ..allocator import ConstExpr
from .eudv import EUDVariable, process_dest
from .vbuf import get_current_custom_varbuffer


# Unused variable don't need to be allocated.
class XVariableTriggerForward(ConstExpr):
    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls, None)

    def __init__(self, initval) -> None:
        super().__init__()
        self._initval = initval

    def Evaluate(self):  # noqa: N802
        evb = get_current_custom_varbuffer()
        try:
            return evb._vdict[self].Evaluate()
        except KeyError:
            vt = evb.create_vartrigger(self, self._initval)
            return vt.Evaluate()


class EUDXVariable(EUDVariable):
    __slots__ = ()

    def __init__(self, epd, modtype, initval, mask=0xFFFFFFFF, /, *, nextptr=0):
        dest = process_dest(epd)
        modifier = ((bt.EncodeModifier(modtype) & 0xFF) << 24) + 0x2D0000
        if not isinstance(initval, int | ConstExpr):
            unproxied = unProxy(initval)
            if not isinstance(unproxied, (int, ConstExpr)):  # noqa: UP038
                raise EPError(_("Invalid initval: {}").format(initval))
            else:
                initval = unproxied
        if not isinstance(nextptr, int | ConstExpr):
            unproxied = unProxy(nextptr)
            if not isinstance(unproxied, (int, ConstExpr)):  # noqa: UP038
                raise EPError(_("Invalid nextptr: {}").format(nextptr))
            else:
                nextptr = unproxied
        if not isinstance(mask, int | ConstExpr):
            unproxied = unProxy(mask)
            if not isinstance(unproxied, (int, ConstExpr)):  # noqa: UP038
                raise EPError(_("Invalid mask: {}").format(mask))
            else:
                mask = unproxied
        args = (mask, dest, initval, modifier, nextptr)

        self._vartrigger = XVariableTriggerForward(args)
        self._varact = self._vartrigger + (8 + 320)
        self._rvalue = False



import os
import sys

sys.path.insert(0, os.path.abspath("..\\"))

import pyximport

pyximport.install()

from eudplib import *

LoadMap("outputmap/basemap/사이비2.scx")


@EUDFunc
def main():
    if EUDWhile()(Always()):
        RunTrigTrigger()
        EUDDoEvents()
    EUDEndWhile()


SaveMap("outputmap/testinlinecode.scx", main)



# Copyright 2024 by zzt (Defender).
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar

from .. import utils as ut
from ..core.rawtrigger.consttype import ConstType
from ..core.rawtrigger.strenc import EncodeUnitOrder
from ..core.rawtrigger.typehint import _UnitOrder
from ..localize import _
from .offsetmap import (
    AnimationMember,
    BoolMember,
    EPDOffsetMap,
    IconMember,
    StatTextMember,
    TechMember,
    UnitOrderMember,
    WeaponMember,
    WordMember,
)


class UnitOrder(EPDOffsetMap, ConstType):
    __slots__ = ()
    label: ClassVar[StatTextMember] = StatTextMember("array", 0x665280)
    useWeaponTargeting: ClassVar[BoolMember] = BoolMember("array", 0x664B00)
    # secondaryOrder: ClassVar[BoolMember] = BoolMember("array", 0x665940)
    # secondary order is for reference only (unused)
    # nonSubunit: ClassVar[BoolMember] = BoolMember("array", 0x665A00)
    # non subunit is for reference only (unused)
    subunitInheritance: ClassVar[BoolMember] = BoolMember("array", 0x664A40)
    "when main unit receives the order, gives it to subunit as well"
    # subunitCanUse: ClassVar[BoolMember] = BoolMember("array", 0x6657C0)
    # subunit can use is for reference only (unused)
    canBeInterrupted: ClassVar[BoolMember] = BoolMember("array", 0x665040)
    canBeLifted: ClassVar[BoolMember] = BoolMember("array", 0x665100)
    "if a movement-type order, sets/clears Lifted (0x20) cunit movement flag"
    canBeQueued: ClassVar[BoolMember] = BoolMember("array", 0x665700)
    disablingKeepsTarget: ClassVar[BoolMember] = BoolMember("array", 0x6651C0)
    "does not clear order target unit when frozen by lockdown/stasis/maelstrom"
    canBeObstructed: ClassVar[BoolMember] = BoolMember("array", 0x6654C0)
    fleeable: ClassVar[BoolMember] = BoolMember("array", 0x664C80)
    "order can be interrupted by unit fleeing from hit reaction"
    # requireMoving: ClassVar[BoolMember] = BoolMember("array", 0x664BC0)
    # require moving is for reference only (unused)
    weapon: ClassVar[WeaponMember] = WeaponMember("array", 0x665880)
    techUsed: ClassVar[TechMember] = TechMember("array", 0x664E00)
    animation: ClassVar[AnimationMember] = AnimationMember("array", 0x664D40)
    icon: ClassVar[IconMember] = IconMember("array", 0x664EC0)
    requirementOffset: ClassVar[WordMember] = WordMember("array", 0x665580)
    obscuredOrder: ClassVar[UnitOrderMember] = UnitOrderMember("array", 0x665400)

    @ut.classproperty
    def range(self):
        return (0, 188, 1)

    @classmethod
    def cast(cls, _from: _UnitOrder):
        if isinstance(_from, ConstType) and not isinstance(_from, cls):
            raise ut.EPError(_('"{}" is not a {}').format(_from, cls.__name__))
        return super().cast(_from)

    def __init__(self, initval: _UnitOrder) -> None:
        super().__init__(EncodeUnitOrder(initval))



from helper import *


@TestInstance
def test_pvariable():
    a = PVariable([5, 5, 5, 5, 5, 5, 5, 5])
    for i in range(7):
        a[i] = 2**i  # (1, 2, 4, 8, 16, 32, 64, 5)

    for i in EUDLoopRange(2, 5):
        a[i] = i * i * i  # (1, 2, 8, 27, 64, 32, 64, 5)

    b = EUDVariable(a)
    c = PVariable.cast(b)

    v_sum = EUDVariable()
    n = EUDVariable(1)
    for i in EUDLoopRange(0, 8):
        v_sum += a[i] * n
        n += 1

    test_equality("PVariable test", [v_sum, a[7]], [1137, 5])

    @EUDTypedFunc([PVariable])
    def sub1(v):
        for i in EUDLoopRange(8):
            v[i] -= 1

    @EUDTypedFunc([PVariable])
    def sub1i(v):
        for i in EUDLoopRange(8):
            v[i] -= 1 + i

    sub1(a)
    test_equality(
        "PVariable -= 1 test",
        [a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7]],
        [0, 1, 7, 26, 63, 31, 63, 4],
    )

    sub1i(a)
    test_equality(
        "PVariable -= var test",
        [a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7]],
        [0xFFFFFFFF, 0xFFFFFFFF, 4, 22, 58, 25, 56, 0xFFFFFFFC],
    )



from helper import *


@TestInstance
def test_bitwise():
    b1100 = EUDVariable(0b1100)
    b1010 = EUDVariable(0b1010)
    test_equality("f_bitand", f_bitand(b1100, b1010), 0b1100 & 0b1010)
    test_equality("f_bitor", f_bitor(b1100, b1010), 0b1100 | 0b1010)
    test_equality("f_bitxor", f_bitxor(b1100, b1010), 0b1100 ^ 0b1010)
    test_equality(
        "f_bitnand", f_bitnand(b1100, b1010), ~(0b1100 & 0b1010) & 0xFFFFFFFF
    )
    test_equality(
        "f_bitnor", f_bitnor(b1100, b1010), ~(0b1100 | 0b1010) & 0xFFFFFFFF
    )
    test_equality(
        "f_bitnxor", f_bitnxor(b1100, b1010), ~(0b1100 ^ 0b1010) & 0xFFFFFFFF
    )
    test_equality("f_bitnot", f_bitnot(b1100), ~0b1100 & 0xFFFFFFFF)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Callable

from ... import core as c
from ...core.eudfunc.trace.tracetool import _f_initstacktrace
from ...eudlib.utilf.userpl import _f_initisusercp, _f_inituserplayerid
from ...localize import _
from ...string.locale import _detect_locale
from ...string.tblprint import _f_initstattext
from ...utils import EPError, ep_assert

_jumper: c.Forward | None = None
_start_functions1: list[Callable] = []
_start_functions2: list[Callable] = []
_has_already_started: int = 0


def has_already_started() -> bool:
    return _has_already_started == 2


def EUDOnStart(func: Callable) -> None:  # noqa: N802
    # FIXME: Callable[[], ...]
    ep_assert(
        _has_already_started < 2,
        "Can't use EUDOnStart here. See https://cafe.naver.com/edac/69262",
    )
    if _has_already_started == 0:
        _start_functions1.append(func)
    else:
        eud_onstart2(func)


def eud_onstart2(func: Callable) -> None:
    _start_functions2.append(func)


def main_starter(mf: Callable) -> c.Forward:
    global _jumper, _has_already_started
    _jumper = c.Forward()

    if c.PushTriggerScope():
        rootstarter = c.NextTrigger()

        # Various initializes
        _detect_locale()
        _f_inituserplayerid()
        _f_initisusercp()
        _f_initstacktrace()
        _f_initstattext()

        for func in _start_functions1:
            func()
        _has_already_started = 1
        start2 = c.Forward()
        c.SetNextTrigger(start2)

        mf_start = c.NextTrigger()
        mf()

        _has_already_started = 2
        if _start_functions2:
            c.PushTriggerScope()
            start2 << c.NextTrigger()
            for func in _start_functions2:
                func()
            c.SetNextTrigger(mf_start)
            c.PopTriggerScope()
        else:
            start2 << mf_start

        c.RawTrigger(nextptr=0x80000000, actions=c.SetNextPtr(_jumper, 0x80000000))
        _jumper << c.RawTrigger(nextptr=rootstarter)

    c.PopTriggerScope()
    _has_already_started = 0

    return _jumper


def EUDDoEvents() -> None:  # noqa: N802
    from ... import memio

    if _jumper is None:
        raise EPError(_("main_starter has not called"))
    oldcp = memio.f_getcurpl()

    _t = c.Forward()
    c.RawTrigger(nextptr=0x80000000, actions=c.SetNextPtr(_jumper, _t))
    _t << c.NextTrigger()

    memio.f_setcurpl(oldcp)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING, Literal

from ... import utils as ut
from ..allocator import (
    ConstExpr,
    GetObjectAddr,
    RegisterCreatePayloadCallback,
    RlocInt_C,
)
from ..eudobj import EUDObject

if TYPE_CHECKING:
    from ..allocator.payload import ObjCollector
    from .eudv import VariableTriggerForward


class EUDVarBuffer(EUDObject):
    """Variable buffer

    72 bytes per variable.
    """

    def __init__(self) -> None:
        super().__init__()

        self._vdict: dict[VariableTriggerForward, ConstExpr] = {}
        self._initvals: list[int | ConstExpr] = []

    def DynamicConstructed(self) -> Literal[True]:  # noqa: N802
        return True

    def Evaluate(self) -> RlocInt_C:  # noqa: N802
        return GetObjectAddr(self) - 4

    def create_vartrigger(self, v, initval) -> ConstExpr:
        ret = self + (72 * len(self._initvals))
        self._initvals.append(initval)
        self._vdict[v] = ret
        return ret

    def create_vartriggers(self, v, initvals) -> ConstExpr:
        ret = self + (72 * len(self._initvals))
        self._initvals.extend(initvals)
        self._vdict[v] = ret
        return ret

    def GetDataSize(self) -> int:  # noqa: N802
        return 2404 + 72 * (len(self._initvals) - 1)

    def CollectDependency(self, emitbuffer: ObjCollector) -> None:  # noqa: N802
        for initval in self._initvals:
            if not isinstance(initval, int):
                emitbuffer.WriteDword(initval)

    def WritePayload(self, emitbuffer) -> None:  # noqa: N802
        for _ in range(4):
            emitbuffer.WriteDword(0)  # nextptr
            emitbuffer.WriteSpace(15)
            emitbuffer.WriteByte(0)  # nocond
            emitbuffer.WriteSpace(52)

        emitbuffer.WriteDword(0)  # nextptr
        emitbuffer.WriteSpace(15)
        emitbuffer.WriteByte(0)  # nocond
        emitbuffer.WriteSpace(16)

        output = bytearray(72)
        output[0:4] = b"\xFF\xFF\xFF\xFF"
        output[24:36] = b"\0\0\x2D\x07\0\0SC\x04\0\0\0"
        output = output * len(self._initvals)

        heads = 0
        for i, initval in enumerate(self._initvals):
            if initval == 0:
                continue

            heade = i * 72 + 20
            if isinstance(initval, int):
                output[heade : heade + 4] = ut.i2b4(initval)
                continue
            emitbuffer.WriteBytes(bytes(output[heads:heade]))
            emitbuffer.WriteDword(initval)
            heads = heade + 4

        emitbuffer.WriteBytes(bytes(output[heads:]))

        emitbuffer.WriteSpace(32)
        emitbuffer.WriteDword(4)  # flags
        emitbuffer.WriteSpace(27)
        emitbuffer.WriteByte(0)  # currentAction

        for _ in range(27):
            emitbuffer.WriteSpace(40)
            emitbuffer.WriteDword(4)  # flags
            emitbuffer.WriteSpace(27)
            emitbuffer.WriteByte(0)  # currentAction


_evb = None


def _register_new_varbuffer() -> None:
    global _evb
    _evb = EUDVarBuffer()


def get_current_varbuffer() -> EUDVarBuffer | None:
    return _evb


RegisterCreatePayloadCallback(_register_new_varbuffer)


class EUDCustomVarBuffer(EUDObject):
    """Customizable Variable buffer

    72 bytes per variable.
    """

    def __init__(self):
        super().__init__()

        self._vdict = {}
        self._5nptrs = deque([], maxlen=5)
        self._actnptr_pairs = []
        self._5acts = deque([], maxlen=5)

    def DynamicConstructed(self):  # noqa: N802
        return True

    def Evaluate(self) -> RlocInt_C:  # noqa: N802
        return GetObjectAddr(self) - 4

    def create_vartrigger(self, v, initval):
        # bitmask, player, #, modifier, nptr
        ret = self + 72 * (len(self._actnptr_pairs) + len(self._5acts))
        if len(self._5acts) == 5:
            act = self._5acts.popleft()
            act.append(initval[4])
            self._actnptr_pairs.append(act)
        else:
            self._5nptrs.append(initval[4])
        self._5acts.append(deque(initval[0:4], maxlen=5))
        if v is not None:
            self._vdict[v] = ret
        return ret

    def create_vartriggers(self, v, initvals):
        ret = self + 72 * (len(self._actnptr_pairs) + len(self._5acts))
        for initval in initvals:
            self.create_vartrigger(None, initval)
        self._vdict[v] = ret
        return ret

    def GetDataSize(self):  # noqa: N802
        return 2404 + 72 * (len(self._actnptr_pairs) + len(self._5acts) - 1)

    def CollectDependency(self, emitbuffer: ObjCollector):  # noqa: N802
        for initval in self._5nptrs:
            if not isinstance(initval, int):
                emitbuffer.WriteDword(initval)
        for initvals in self._actnptr_pairs:
            for initval in initvals:
                if not isinstance(initval, int):
                    emitbuffer.WriteDword(initval)
        for initvals in self._5acts:
            for initval in initvals:
                if not isinstance(initval, int):
                    emitbuffer.WriteDword(initval)

    def WritePayload(self, emitbuffer) -> None:  # noqa: N802
        output = bytearray(72 * (len(self._actnptr_pairs) + len(self._5acts)))

        for i in range(5):
            try:
                emitbuffer.WriteDword(self._5nptrs[i])  # nextptr
            except IndexError:
                emitbuffer.WriteDword(0)
            emitbuffer.WriteSpace(15)
            emitbuffer.WriteByte(0)  # nocond
            if i < 4:
                emitbuffer.WriteSpace(52)
            else:
                emitbuffer.WriteSpace(16)

        for i in range(len(self._actnptr_pairs) + len(self._5acts)):
            # 'preserve rawtrigger'
            output[72 * i : 72 * i + 4] = b"\xFF\xFF\xFF\xFF"
            output[72 * i + 24 : 72 * i + 28] = b"\0\0\x2D\x07"
            output[72 * i + 28 : 72 * i + 32] = b"\0\0SC"
            output[72 * i + 32 : 72 * i + 36] = b"\x04\0\0\0"

        heads = 0

        from itertools import chain

        # bitmask, player, initval, modifier, nptr
        offsets = (0, 16, 20, 24, 36)
        initials = (0xFFFFFFFF, 0, 0, 0x072D0000, 0)
        for i, initvals in enumerate(chain(self._actnptr_pairs, self._5acts)):
            for initval, offset, initial in zip(initvals, offsets, initials):
                heade = 72 * i + offset
                if initval == initial:
                    continue
                elif isinstance(initval, int):
                    output[heade : heade + 4] = ut.i2b4(initval)
                    continue
                emitbuffer.WriteBytes(bytes(output[heads:heade]))
                emitbuffer.WriteDword(initval)
                heads = 72 * i + offset + 4

        emitbuffer.WriteBytes(bytes(output[heads:]))

        emitbuffer.WriteSpace(32)
        emitbuffer.WriteDword(4)  # flags
        emitbuffer.WriteSpace(27)
        emitbuffer.WriteByte(0)  # currentAction

        for _ in range(27):
            emitbuffer.WriteSpace(40)
            emitbuffer.WriteDword(4)  # flags
            emitbuffer.WriteSpace(27)
            emitbuffer.WriteByte(0)  # currentAction


_ecvb = None


def _register_new_custom_varbuffer():
    global _ecvb
    _ecvb = EUDCustomVarBuffer()


def get_current_custom_varbuffer():
    return _ecvb


RegisterCreatePayloadCallback(_register_new_custom_varbuffer)



# Copyright 2024 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Callable

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from . import modcurpl as cp

cpcache = c.curpl.GetCPCache()
# consecutive_table[shift][endswith][startswith]
consecutive_table: dict[int, dict[int, dict[int, c.RawTrigger]]] = {}

c.PushTriggerScope()
copy_ret = c.SetDeaths(0, c.SetTo, 0, 0)
read_end_common = c.RawTrigger(
    nextptr=0,
    actions=[
        copy_ret,
        c.SetMemory(copy_ret + 20, c.SetTo, 0),
    ],
)
c.PopTriggerScope()


def _is_consecutive(n: int) -> bool:
    # right shift by index of least significant bit (remove trailing zeros)
    n >>= (n & -n).bit_length() - 1
    return (n & (n + 1)) == 0


def guess_ordering(bitmask: int, shift: int) -> bool:
    lsb_index = (bitmask & -bitmask).bit_length() - 1
    if lsb_index == -shift:
        return True
    elif shift == 0:
        return False
    else:
        return True


def signed_shift(a: int, b: int):
    if b >= 0:
        return a << b
    else:
        return a >> -b


def _insert(
    bitmask: int, shift: int, is_decreasing: bool | None = None
) -> c.RawTrigger:
    ut.ep_assert(
        _is_consecutive(bitmask), f"Not a consecutive bit pattern: {bitmask:x}"
    )
    ut.ep_assert(
        is_decreasing in (True, False, None), f"Invalid ordering: {is_decreasing}"
    )
    if is_decreasing is None:
        is_decreasing = guess_ordering(bitmask, shift)

    msb_index: int = bitmask.bit_length() - 1
    lsb_index = (bitmask & -bitmask).bit_length() - 1
    if is_decreasing:
        startswith, endswith = msb_index, lsb_index
    else:
        startswith, endswith = lsb_index, msb_index

    try:
        shift_table = consecutive_table[shift]
    except KeyError:
        shift_table = {}
        consecutive_table[shift] = shift_table
    try:
        trig_table = shift_table[endswith]
    except KeyError:
        trig_table = {}
        shift_table[endswith] = trig_table
    try:
        start_trig = trig_table[startswith]
    except KeyError:
        c.PushTriggerScope()
        step = 1 if is_decreasing else -1
        for x in range(endswith, startswith + step, step):
            if x not in trig_table:
                trig = c.RawTrigger(
                    nextptr=read_end_common
                    if x == endswith
                    else trig_table[x - step],
                    conditions=c.DeathsX(cp.CP, c.AtLeast, 1, 0, 1 << x),
                    actions=c.SetMemory(
                        copy_ret + 20, c.Add, signed_shift(1, x + shift)
                    ),
                )
                trig_table[x] = trig
        c.PopTriggerScope()
        start_trig = trig_table[startswith]

    return start_trig


def _get(bitmask: int, shift: int) -> c.RawTrigger:
    ut.ep_assert(
        _is_consecutive(bitmask), f"Not a consecutive bit pattern: {bitmask:x}"
    )
    msb_index = bitmask.bit_length() - 1
    lsb_index = (bitmask & -bitmask).bit_length() - 1

    shift_table = consecutive_table[shift]
    try:
        trig_table = shift_table[lsb_index]
    except KeyError:
        trig_table = shift_table[msb_index]
        return trig_table[lsb_index]
    else:
        return trig_table[msb_index]


def _insert_or_get(
    bitmask: int, shift: int, is_decreasing: bool | None = None
) -> c.RawTrigger:
    try:
        return _get(bitmask, shift)
    except KeyError:
        return _insert(bitmask, shift, is_decreasing)


def _cp_caller(readtrg: c.RawTrigger, *, _operations=None) -> Callable:
    def f(cpo=0, *, ret=None):
        if ret is None:
            ret = c.EUDVariable()
            ret.makeR()
        else:
            ret = ret[0]
        if c.IsEUDVariable(ret):
            retepd = ut.EPD(ret.getValueAddr())
        else:
            retepd = ret

        operations = []
        if not (isinstance(cpo, int) and cpo == 0):
            operations.append((ut.EPD(0x6509B0), c.Add, cpo))

        nexttrg = c.Forward()
        operations.extend(
            (
                (ut.EPD(read_end_common) + 1, c.SetTo, nexttrg),
                (ut.EPD(copy_ret) + 4, c.SetTo, retepd),
            )
        )
        if _operations:
            operations.extend(_operations)
        c.NonSeqCompute(operations)
        c.SetNextTrigger(readtrg)
        nexttrg << c.NextTrigger()

        if not (isinstance(cpo, int) and cpo == 0):
            cs.DoActions(c.SetMemory(0x6509B0, c.Add, -cpo))

        return ret

    return f


def _epd_caller(readtrg: c.RawTrigger, *, _operations=None) -> Callable:
    def f(epd, *, ret=None):
        if ret is None:
            ret = c.EUDVariable()
            ret.makeR()
        else:
            ret = ret[0]
        if c.IsEUDVariable(ret):
            retepd = ut.EPD(ret.getValueAddr())
        else:
            retepd = ret

        nexttrg = c.Forward()
        operations = [
            (ut.EPD(read_end_common) + 1, c.SetTo, cpcache.GetVTable()),
            (ut.EPD(cpcache.GetVTable()) + 1, c.SetTo, nexttrg),
            (ut.EPD(cpcache.getDestAddr()), c.SetTo, ut.EPD(0x6509B0)),
            (ut.EPD(copy_ret) + 4, c.SetTo, retepd),
            (ut.EPD(0x6509B0), c.SetTo, epd),
        ]
        if _operations:
            operations.extend(_operations)
        c.NonSeqCompute(operations)
        c.SetNextTrigger(readtrg)
        nexttrg << c.NextTrigger()

        return ret

    return f


_insert(0xFFFFFFFF, 0, True)
_insert(0xFFFFFF00, 0, False)
_br = (
    _get(0xFF, 0),
    _insert(0x0000FF00, -8, True),
    _insert(0x00FF0000, -16, True),
    _insert(0xFF000000, -24, True),
)



from helper import *


@TestInstance
def test_binsearch():
    n = EUDVariable()

    # Basic test
    n << 1000
    p1 = EUDBinaryMin(lambda x: x * x >= n, 0, 0xFFFF)
    p2 = EUDBinaryMax(lambda x: x * x <= n, 0, 0xFFFF)
    test_equality("Binary search - Square root test", [p1, p2], [32, 31])

    # Specific range test
    @EUDFunc
    def comp1(x):
        if EUDIf()(x <= 30):
            EUDReturn(0)
        if EUDElseIf()(x >= 70):
            EUDReturn(0)
        if EUDElseIf()(x <= 40):
            EUDReturn(1)
        if EUDElse()():
            EUDReturn(0)
        EUDEndIf()

    p3 = EUDBinaryMax(comp1, 31, 69)

    @EUDFunc
    def comp1(x):
        if EUDIf()(x <= 30):
            EUDReturn(0)
        if EUDElseIf()(x >= 70):
            EUDReturn(0)
        if EUDElseIf()(x >= 40):
            EUDReturn(1)
        if EUDElse()():
            EUDReturn(0)
        EUDEndIf()

    p4 = EUDBinaryMin(comp1, 31, 69)

    test_equality("Binary search - bounded range test", [p3, p4], [40, 40])



# Copyright 2024 by zzt (Defender).
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar

from .. import utils as ut
from ..core.rawtrigger.consttype import ConstType
from ..core.rawtrigger.strenc import EncodeImage
from ..core.rawtrigger.typehint import _Image
from ..localize import _
from .offsetmap import BoolMember, DrawingFunctionMember, EPDOffsetMap, IscriptMember


class Image(EPDOffsetMap, ConstType):
    __slots__ = ()
    # Read only data skipped
    # grpFile: ClassVar[ArrayMember] = ArrayMember(0x668AA0, Mk.DWORD)
    isTurnable: ClassVar[BoolMember] = BoolMember("array", 0x66E860)
    graphicTurn = isTurnable
    isClickable: ClassVar[BoolMember] = BoolMember("array", 0x66C150)
    useFullIscript: ClassVar[BoolMember] = BoolMember("array", 0x66D4D8)
    drawIfCloaked: ClassVar[BoolMember] = BoolMember("array", 0x667718)
    drawingFunction: ClassVar[DrawingFunctionMember] = DrawingFunctionMember(
        "array", 0x669E28
    )
    # FIXME: Add UnsupportedMember?
    # remapping: ClassVar[ArrayMember] = ArrayMember(0x669A40, Mk.BYTE)
    # Remapping table is skipped because it doesn't work in SC:R
    iscript: ClassVar[IscriptMember] = IscriptMember("array", 0x66EC48)
    # shieldsOverlay: ClassVar[ArrayMember] = ArrayMember(0x66C538, Mk.DWORD)
    # attackOverlay: ClassVar[ArrayMember] = ArrayMember(0x66B1B0, Mk.DWORD)
    # damageOverlay: ClassVar[ArrayMember] = ArrayMember(0x66A210, Mk.DWORD)
    # specialOverlay: ClassVar[ArrayMember] = ArrayMember(0x667B00, Mk.DWORD)
    # landingDustOverlay: ClassVar[ArrayMember] = ArrayMember(0x666778, Mk.DWORD)
    # liftOffDustOverlay: ClassVar[ArrayMember] = ArrayMember(0x66D8C0, Mk.DWORD)

    @ut.classproperty
    def range(self):
        return (0, 998, 1)

    @classmethod
    def cast(cls, _from: _Image):
        if isinstance(_from, ConstType) and not isinstance(_from, cls):
            raise ut.EPError(_('"{}" is not a {}').format(_from, cls.__name__))
        return super().cast(_from)

    def __init__(self, initval: _Image) -> None:
        super().__init__(EncodeImage(initval))



# Print possible path to each states

import re

re_state = re.compile(r"^State (\d+):$")
re_shift = re.compile(r"^ +(\w+) shift  (\d+)$")

with open("epparser.out") as file:
    lines = file.readlines()

# Get graph
graph: dict[int, list[tuple[str, int]]] = {}
_current_state: int = 0
for line in lines:
    r_st = re_state.match(line)
    if r_st:
        _current_state = int(r_st.group(1))
        graph[_current_state] = []
    else:
        r_sh = re_shift.match(line)
        if r_sh:
            token = r_sh.group(1)
            _state_to = int(r_sh.group(2))
            graph[_current_state].append((token, _state_to))

# Find path
_path_map = {}
q = [("", 0)]

while q:
    _prev_path, _current_state = q.pop()
    for token, _next_state in graph[_current_state]:
        if _next_state in _path_map:
            continue
        path = f"{_prev_path} {token}"
        _path_map[_next_state] = path
        q.append((path, _next_state))

keys = list(_path_map.keys())
keys.sort()
for k in keys:
    print("%5d :%s" % (k, _path_map[k]))



from helper import *


@TestInstance
def test_vdefval():
    a = EUDVariable(5)
    test_equality("Variable with default value", a, 5)


@TestInstance
def test_vmixedtrg():
    # VMixed actions
    a = EUDVariable()

    a << 0
    DoActions(SetDeaths(a, SetTo, EPD(a), 0))
    a << 1
    DoActions(SetDeaths(a, SetTo, f_mul(a, 30), 0))
    a << 2
    DoActions(SetDeaths(a, SetTo, a - 50, 0))
    DoActions(SetDeaths(3, SetTo, 123, a), SetDeaths(a, SetTo, a, a))

    test_assert(
        "Variable mixed trigger test",
        [
            Deaths(0, Exactly, 0x3FE9D727, 0),  # EPD(0) == 0x3fe9d727
            Deaths(1, Exactly, 30, 0),
            Deaths(2, Exactly, -48, 0),
            Deaths(3, Exactly, 123, 2),
            Deaths(2, Exactly, 2, 2),
        ],
    )

    DoActions(SetDeaths(AllPlayers, SetTo, 0, 0), SetDeaths(AllPlayers, SetTo, 0, 2))
    a <<= 1
    a -= Db(4)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .epscompile import EPS_SetDebug, epsCompile
from .epsimp import EPSLoader

__all__ = [
    "EPS_SetDebug",
    "EPSLoader",
    "epsCompile",
]



# Copyright 2020 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import ByteString

from ... import utils as ut
from ...core.allocator.pbuffer import Payload
from ...core.mapdata.chktok import CHK
from ...core.mapdata.stringmap import get_string_map, get_string_section_name
from ...localize import _
from ...trigtrg import trigtrg as tt

""" Stage 1 : works in TRIG section.
- Check if EUD is available. If not, defeat
- Initialize stage 2 & execute it
- Modifies TRIG rawtrigger's nextptr. Modification is fixed in stage 3.
"""
trglist: list[bytes] = []


def _Trigger(  # noqa: N802
    *,
    # FIXME: Cannot determine type of "AllPlayers" [has-type]
    players: list[tt.Player] = [17],
    conditions: list[bytes] | bytes = [],
    actions: list[bytes] | bytes = [],
) -> None:
    global trglist
    trglist.append(tt.Trigger(players, conditions, actions))


def initialize_payload(
    chkt: CHK, payload: Payload, mrgndata: ByteString | None = None
) -> None:
    strmap = get_string_map()
    if strmap is None:
        raise ut.EPError(_("Must use LoadMap first"))
    str_section = strmap.save_tbl()
    str_padding = -len(str_section) & 3
    payload_offset = 0x191943C8 + len(str_section) + str_padding

    orig_payload = payload.data
    new_payload = bytearray(orig_payload)

    for prt in payload.prttable:
        v = ut.b2i4(new_payload[prt : prt + 4])
        v += payload_offset // 4
        new_payload[prt : prt + 4] = ut.i2b4(v)
    for ort in payload.orttable:
        v = ut.b2i4(new_payload[ort : ort + 4])
        v += payload_offset
        new_payload[ort : ort + 4] = ut.i2b4(v)

    str_section = str_section + bytes(str_padding) + new_payload
    chkt.setsection(get_string_section_name(), str_section)

    if mrgndata is not None:
        orig_mrgn = chkt.getsection("MRGN")
        mrgn_section = []
        for i in range(0, len(orig_mrgn), 4):
            if i % 20 == 16:  # remove name
                orig = ut.b2i2(orig_mrgn, i + 2) << 16
            else:
                orig = ut.b2i4(orig_mrgn, i)
            obfus = ut.b2i4(mrgndata, i)
            if obfus > orig:
                orig += 1 << 32
            mrgn_section.append(ut.i2b4(orig - obfus))
        chkt.setsection("MRGN", b"".join(mrgn_section))

    ##############
    # TRIG SECTION
    ##############
    trglist.clear()

    # pts[player].lasttrigger->next = value(strs) + strsled_offset

    pts = 0x51A284
    curpl = 0x6509B0

    for player in ut._rand_lst(range(8)):
        triggerend = ~(pts + player * 12)

        _Trigger(
            players=[player],
            actions=[
                tt.SetMemory(curpl, tt.SetTo, ut.EPD(pts + 12 * player)),
                tt.SetDeaths(9, tt.SetTo, triggerend, 0),  # Used in stage 2
            ],
        )

    # read pts[player].lasttrigger
    for e in ut._rand_lst(range(2, 32)):
        _Trigger(
            conditions=tt.DeathsX(tt.CurrentPlayer, tt.AtLeast, 1, 0, 2**e),
            actions=tt.SetDeaths(11, tt.Add, 2**e, 0),
        )

    # apply to curpl
    _Trigger(
        actions=[
            tt.SetDeaths(10, tt.SetTo, ut.EPD(4), 0),
            tt.SetMemory(curpl, tt.SetTo, ut.EPD(4)),
        ]
    )
    for e in ut._rand_lst(range(2, 32)):
        _Trigger(
            conditions=tt.DeathsX(11, tt.AtLeast, 1, 0, 2**e),
            actions=[
                # tt.SetDeaths(11, tt.Subtract, 2 ** e, 0),
                # used for nextptr recovery in stage 3
                tt.SetDeaths(10, tt.Add, 2 ** (e - 2), 0),
                tt.SetMemory(curpl, tt.Add, 2 ** (e - 2)),
            ],
        )

    # now curpl = EPD(value(ptsprev) + 4)
    # value(EPD(value(ptsprev) + 4)) = strs + payload_offset
    # copy_deaths(tt.EPD(strs), tt.CurrentPlayer, False, strsled_offset)
    _Trigger(
        actions=[
            tt.SetDeaths(tt.CurrentPlayer, tt.SetTo, payload_offset, 0),
            tt.SetDeaths(11, tt.SetTo, 0, 0),
        ]
    )

    # Done!
    trigdata = b"".join(trglist)

    # Stage 1 created

    # -------

    # Previous rawtrigger datas

    oldtrigraw = chkt.getsection("TRIG")
    oldtrigs = [oldtrigraw[i : i + 2400] for i in range(0, len(oldtrigraw), 2400)]
    proc_trigs = []

    # Collect only enabled triggers
    for trig in oldtrigs:
        trig = bytearray(trig)
        flag = ut.b2i4(trig, 320 + 2048)
        if flag & 8:  # Trigger already disabled
            pass

        flag ^= 8  # Disable it temporarily. It will be re-enabled at stage 3
        trig[320 + 2048 : 320 + 2048 + 4] = ut.i2b4(flag)
        proc_trigs.append(bytes(trig))

    oldtrigraw_processed = b"".join(proc_trigs)
    chkt.setsection("TRIG", trigdata + oldtrigraw_processed)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.
from collections.abc import Iterator
from typing import Literal

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...core.eudfunc.eudf import _EUDPredefineReturn
from ...localize import _
from ...memio import f_getcurpl, f_setcurpl
from ...scdata import TrgPlayer


@_EUDPredefineReturn(2, 3)
@c.EUDTypedFunc([TrgPlayer], [None])
def f_playerexist(player):
    """Check if player has not left the game.

    :returns: 1 if player exists, 0 if not.
    """
    ret = f_playerexist._frets[0]
    pts = 0x51A280

    cs.EUDSwitch(player)
    block = ut.EUDGetLastBlockOfName("swblock")[1]
    block["_actions"].append(ret.SetNumber(1))
    for p in ut._rand_lst(range(8)):
        if cs.EUDSwitchCase()(p):
            c.RawTrigger(
                nextptr=block["swend"],
                conditions=c.Memory(
                    pts + p * 12 + 8, c.Exactly, ~(pts + p * 12 + 4)
                ),
                actions=ret.SetNumber(0),
            )

    if cs.EUDSwitchDefault()():
        ret << 0
    cs.EUDEndSwitch()
    # return ret


# --------


def EUDLoopPlayer(  # noqa: N802
    ptype: Literal[
        "Zerg",
        "Terran",
        "Protoss",
        "Independent",
        "Neutral",
        "User selectable",
        "Inactive",
        "Human",
        "Human",
        "Starcraft Campaign Editor",
        "",
    ]
    | None = "Human",
    force: int | TrgPlayer | None = None,
    race: Literal["Unused", "Rescuable", "Computer", "Human", "Neutral", ""]
    | None = None,
) -> Iterator[TrgPlayer]:
    def encode_force(f) -> int:
        if isinstance(f, TrgPlayer) and isinstance(f._value, int):
            ut.ep_assert(18 <= f._value <= 21, _("{} is not force").format(f))
            f = f._value
        ut.ep_assert(isinstance(f, int), _("{} is not force").format(f))
        if 18 <= f <= 21:
            f -= 18
        ut.ep_assert(0 <= f <= 3, _("{} is not force").format(f))
        return f

    plist = []
    for p in range(8):
        pinfo = c.GetPlayerInfo(p)
        if (
            (not ptype or pinfo.typestr == ptype)
            and (not force or pinfo.force == encode_force(force))
            and (not race or pinfo.racestr == race)
        ):
            plist.append(p)

    if not plist:
        e = []
        e.append(_("No player met condition for input map settings:"))
        if ptype is not None:
            e.append(_(" ptype {}").format(ptype))
        if force is not None:
            e.append(_(" force {}").format(force))
        if race is not None:
            e.append(_(" race {}").format(race))
        e.append("\n")
        e.append(_("Check out whether Start Locations are placed correctly."))
        raise ut.EPError("".join(e))
    start, end = min(plist), max(plist)

    ut.EUDCreateBlock("loopplayerblock", None)
    v = c.EUDVariable()
    v << start
    if cs.EUDWhile()(v <= end):
        for i in range(start, end):
            if i not in plist:
                cs.EUDContinueIf(v == i)
        cs.EUDContinueIfNot(f_playerexist(v))
        yield TrgPlayer.cast(v)
        cs.EUDSetContinuePoint()
        v += 1
    cs.EUDEndWhile()
    ut.EUDPopBlock("loopplayerblock")


# -------


def EUDPlayerLoop():  # noqa: N802
    def _footer():
        block = {"origcp": f_getcurpl(), "playerv": c.EUDVariable()}
        playerv = block["playerv"]

        playerv << 0
        cs.EUDWhile()(playerv <= 7)
        cs.EUDContinueIfNot(f_playerexist(playerv))
        f_setcurpl(playerv)

        ut.EUDCreateBlock("ploopblock", block)
        return True

    return cs.CtrlStruOpener(_footer)


def EUDEndPlayerLoop():  # noqa: N802
    block = ut.EUDPopBlock("ploopblock")[1]
    playerv = block["playerv"]
    origcp = block["origcp"]

    if not cs.EUDIsContinuePointSet():
        cs.EUDSetContinuePoint()

    playerv += 1
    cs.EUDEndWhile()
    f_setcurpl(origcp)



from __future__ import annotations

import sys
from contextlib import redirect_stderr, redirect_stdout
from io import StringIO

import pytest

from pybind11_tests import iostream as m


def test_captured(capsys):
    msg = "I've been redirected to Python, I hope!"
    m.captured_output(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr

    m.captured_err(msg)
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr == msg


def test_captured_large_string(capsys):
    # Make this bigger than the buffer used on the C++ side: 1024 chars
    msg = "I've been redirected to Python, I hope!"
    msg = msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_captured_utf8_2byte_offset0(capsys):
    msg = "\u07ff"
    msg = "" + msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_captured_utf8_2byte_offset1(capsys):
    msg = "\u07ff"
    msg = "1" + msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_captured_utf8_3byte_offset0(capsys):
    msg = "\uffff"
    msg = "" + msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_captured_utf8_3byte_offset1(capsys):
    msg = "\uffff"
    msg = "1" + msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_captured_utf8_3byte_offset2(capsys):
    msg = "\uffff"
    msg = "12" + msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_captured_utf8_4byte_offset0(capsys):
    msg = "\U0010ffff"
    msg = "" + msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_captured_utf8_4byte_offset1(capsys):
    msg = "\U0010ffff"
    msg = "1" + msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_captured_utf8_4byte_offset2(capsys):
    msg = "\U0010ffff"
    msg = "12" + msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_captured_utf8_4byte_offset3(capsys):
    msg = "\U0010ffff"
    msg = "123" + msg * (1024 // len(msg) + 1)

    m.captured_output_default(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_guard_capture(capsys):
    msg = "I've been redirected to Python, I hope!"
    m.guard_output(msg)
    stdout, stderr = capsys.readouterr()
    assert stdout == msg
    assert not stderr


def test_series_captured(capture):
    with capture:
        m.captured_output("a")
        m.captured_output("b")
    assert capture == "ab"


def test_flush(capfd):
    msg = "(not flushed)"
    msg2 = "(flushed)"

    with m.ostream_redirect():
        m.noisy_function(msg, flush=False)
        stdout, stderr = capfd.readouterr()
        assert not stdout

        m.noisy_function(msg2, flush=True)
        stdout, stderr = capfd.readouterr()
        assert stdout == msg + msg2

        m.noisy_function(msg, flush=False)

    stdout, stderr = capfd.readouterr()
    assert stdout == msg


def test_not_captured(capfd):
    msg = "Something that should not show up in log"
    stream = StringIO()
    with redirect_stdout(stream):
        m.raw_output(msg)
    stdout, stderr = capfd.readouterr()
    assert stdout == msg
    assert not stderr
    assert not stream.getvalue()

    stream = StringIO()
    with redirect_stdout(stream):
        m.captured_output(msg)
    stdout, stderr = capfd.readouterr()
    assert not stdout
    assert not stderr
    assert stream.getvalue() == msg


def test_err(capfd):
    msg = "Something that should not show up in log"
    stream = StringIO()
    with redirect_stderr(stream):
        m.raw_err(msg)
    stdout, stderr = capfd.readouterr()
    assert not stdout
    assert stderr == msg
    assert not stream.getvalue()

    stream = StringIO()
    with redirect_stderr(stream):
        m.captured_err(msg)
    stdout, stderr = capfd.readouterr()
    assert not stdout
    assert not stderr
    assert stream.getvalue() == msg


def test_multi_captured(capfd):
    stream = StringIO()
    with redirect_stdout(stream):
        m.captured_output("a")
        m.raw_output("b")
        m.captured_output("c")
        m.raw_output("d")
    stdout, stderr = capfd.readouterr()
    assert stdout == "bd"
    assert stream.getvalue() == "ac"


def test_dual(capsys):
    m.captured_dual("a", "b")
    stdout, stderr = capsys.readouterr()
    assert stdout == "a"
    assert stderr == "b"


def test_redirect(capfd):
    msg = "Should not be in log!"
    stream = StringIO()
    with redirect_stdout(stream):
        m.raw_output(msg)
    stdout, stderr = capfd.readouterr()
    assert stdout == msg
    assert not stream.getvalue()

    stream = StringIO()
    with redirect_stdout(stream), m.ostream_redirect():
        m.raw_output(msg)
    stdout, stderr = capfd.readouterr()
    assert not stdout
    assert stream.getvalue() == msg

    stream = StringIO()
    with redirect_stdout(stream):
        m.raw_output(msg)
    stdout, stderr = capfd.readouterr()
    assert stdout == msg
    assert not stream.getvalue()


def test_redirect_err(capfd):
    msg = "StdOut"
    msg2 = "StdErr"

    stream = StringIO()
    with redirect_stderr(stream), m.ostream_redirect(stdout=False):
        m.raw_output(msg)
        m.raw_err(msg2)
    stdout, stderr = capfd.readouterr()
    assert stdout == msg
    assert not stderr
    assert stream.getvalue() == msg2


def test_redirect_both(capfd):
    msg = "StdOut"
    msg2 = "StdErr"

    stream = StringIO()
    stream2 = StringIO()
    with redirect_stdout(stream), redirect_stderr(stream2), m.ostream_redirect():
        m.raw_output(msg)
        m.raw_err(msg2)
    stdout, stderr = capfd.readouterr()
    assert not stdout
    assert not stderr
    assert stream.getvalue() == msg
    assert stream2.getvalue() == msg2


@pytest.mark.skipif(sys.platform.startswith("emscripten"), reason="Requires threads")
def test_threading():
    with m.ostream_redirect(stdout=True, stderr=False):
        # start some threads
        threads = []

        # start some threads
        for _j in range(20):
            threads.append(m.TestThread())

        # give the threads some time to fail
        threads[0].sleep()

        # stop all the threads
        for t in threads:
            t.stop()

        for t in threads:
            t.join()

        # if a thread segfaults, we don't get here
        assert True
